from functools import partial

import gdsfactory as gf

from gf180mcu.layers import LAYER, layer
from gf180mcu.via_generator import via_generator

dn_rect = partial(gf.components.rectangle, layer=LAYER.dnwell)


@gf.cell
def pcmpgr_gen(dn_rect=dn_rect, grw: float = 0.36) -> gf.Component:
    """Return deepnwell guardring.

    Args:
        dn_rect : deepnwell polygon.
        grw : guardring width.

    """
    c = gf.Component()

    dn_rect = gf.get_component(dn_rect)

    comp_pp_enc: float = 0.16
    con_size = 0.22
    con_sp = 0.28
    con_comp_enc = 0.07
    pcmpgr_enc_dn = 2.5

    c_temp_gr = gf.Component("temp_store guard ring")
    rect_pcmpgr_in = c_temp_gr.add_ref(
        gf.components.rectangle(
            size=(
                (dn_rect.dxmax - dn_rect.dxmin) + 2 * pcmpgr_enc_dn,
                (dn_rect.dymax - dn_rect.dymin) + 2 * pcmpgr_enc_dn,
            ),
            layer=layer["comp"],
        )
    )
    rect_pcmpgr_in.dmove((dn_rect.dxmin - pcmpgr_enc_dn, dn_rect.dymin - pcmpgr_enc_dn))
    rect_pcmpgr_out = c_temp_gr.add_ref(
        gf.components.rectangle(
            size=(
                (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) + 2 * grw,
                (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) + 2 * grw,
            ),
            layer=layer["comp"],
        )
    )
    rect_pcmpgr_out.dmove((rect_pcmpgr_in.dxmin - grw, rect_pcmpgr_in.dymin - grw))
    c.add_ref(
        gf.boolean(
            A=rect_pcmpgr_out,
            B=rect_pcmpgr_in,
            operation="A-B",
            layer=layer["comp"],
        )
    )  # guardring bulk

    psdm_in = c_temp_gr.add_ref(
        gf.components.rectangle(
            size=(
                (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) - 2 * comp_pp_enc,
                (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) - 2 * comp_pp_enc,
            ),
            layer=layer["pplus"],
        )
    )
    psdm_in.dmove(
        (
            rect_pcmpgr_in.dxmin + comp_pp_enc,
            rect_pcmpgr_in.dymin + comp_pp_enc,
        )
    )
    psdm_out = c_temp_gr.add_ref(
        gf.components.rectangle(
            size=(
                (rect_pcmpgr_out.dxmax - rect_pcmpgr_out.dxmin) + 2 * comp_pp_enc,
                (rect_pcmpgr_out.dymax - rect_pcmpgr_out.dymin) + 2 * comp_pp_enc,
            ),
            layer=layer["pplus"],
        )
    )
    psdm_out.dmove(
        (
            rect_pcmpgr_out.dxmin - comp_pp_enc,
            rect_pcmpgr_out.dymin - comp_pp_enc,
        )
    )
    c.add_ref(
        gf.boolean(A=psdm_out, B=psdm_in, operation="A-B", layer=layer["pplus"])
    )  # pplus_draw

    # generatingg contacts

    c.add_ref(
        via_generator(
            x_range=(
                rect_pcmpgr_in.dxmin + con_size,
                rect_pcmpgr_in.dxmax - con_size,
            ),
            y_range=(rect_pcmpgr_out.dymin, rect_pcmpgr_in.dymin),
            via_enclosure=(con_comp_enc, con_comp_enc),
            via_layer=layer["contact"],
            via_size=(con_size, con_size),
            via_spacing=(con_sp, con_sp),
        )
    )  # bottom contact

    c.add_ref(
        via_generator(
            x_range=(
                rect_pcmpgr_in.dxmin + con_size,
                rect_pcmpgr_in.dxmax - con_size,
            ),
            y_range=(rect_pcmpgr_in.dymax, rect_pcmpgr_out.dymax),
            via_enclosure=(con_comp_enc, con_comp_enc),
            via_layer=layer["contact"],
            via_size=(con_size, con_size),
            via_spacing=(con_sp, con_sp),
        )
    )  # upper contact

    c.add_ref(
        via_generator(
            x_range=(rect_pcmpgr_out.dxmin, rect_pcmpgr_in.dxmin),
            y_range=(
                rect_pcmpgr_in.dymin + con_size,
                rect_pcmpgr_in.dymax - con_size,
            ),
            via_enclosure=(con_comp_enc, con_comp_enc),
            via_layer=layer["contact"],
            via_size=(con_size, con_size),
            via_spacing=(con_sp, con_sp),
        )
    )  # right contact

    c.add_ref(
        via_generator(
            x_range=(rect_pcmpgr_in.dxmax, rect_pcmpgr_out.dxmax),
            y_range=(
                rect_pcmpgr_in.dymin + con_size,
                rect_pcmpgr_in.dymax - con_size,
            ),
            via_enclosure=(con_comp_enc, con_comp_enc),
            via_layer=layer["contact"],
            via_size=(con_size, con_size),
            via_spacing=(con_sp, con_sp),
        )
    )  # left contact

    comp_m1_in = c_temp_gr.add_ref(
        gf.components.rectangle(
            size=(rect_pcmpgr_in.dxsize, rect_pcmpgr_in.dysize),
            layer=layer["metal1"],
        )
    )

    comp_m1_out = c_temp_gr.add_ref(
        gf.components.rectangle(
            size=(
                (comp_m1_in.dxsize) + 2 * grw,
                (comp_m1_in.dysize) + 2 * grw,
            ),
            layer=layer["metal1"],
        )
    )
    comp_m1_out.dmove((rect_pcmpgr_in.dxmin - grw, rect_pcmpgr_in.dymin - grw))
    c.add_ref(
        gf.boolean(
            A=rect_pcmpgr_out,
            B=rect_pcmpgr_in,
            operation="A-B",
            layer=layer["metal1"],
        )
    )  # metal1 guardring

    return c


if __name__ == "__main__":
    c = pcmpgr_gen()
    c.show()



import gdsfactory as gf
import numpy as np
from gdsfactory.typings import Float2, LayerSpec

from gf180mcu.layers import layer
from gf180mcu.via_generator import via_generator, via_stack


@gf.cell
def cap_mos_inst(
    lc: float = 0.1,
    wc: float = 0.1,
    cmp_w: float = 0.1,
    con_w: float = 0.1,
    pl_l: float = 0.1,
    cmp_ext: float = 0.1,
    pl_ext: float = 0.1,
    implant_layer: LayerSpec = layer["nplus"],
    implant_enc: Float2 = (0.1, 0.1),
    label: bool = False,
    g_label: str = "",
) -> gf.Component:
    """Returns mos cap simple instance.

    Args:
        lc : length of mos_cap.
        ws : width of mos_cap.
        cmp_w : width of layer["comp"].
        con_w : min width of comp contain contact.
        pl_l : length od layer["poly2"].
        cmp_ext : comp extension beyond poly2.
        pl_ext : poly2 extension beyond comp.
        implant_layer : Layer of implant [nplus,pplus].
        implant_enc : enclosure of implant_layer to comp.
        label : 1 to add labels.
        g_label : gate label.

    """
    c_inst = gf.Component()

    cmp = c_inst.add_ref(gf.components.rectangle(size=(cmp_w, wc), layer=layer["comp"]))

    cap_mk = c_inst.add_ref(
        gf.components.rectangle(
            size=(cmp.dxsize, cmp.dysize), layer=layer["mos_cap_mk"]
        )
    )
    cap_mk.dxmin = cmp.dxmin
    cap_mk.dymin = cmp.dymin

    if (column_pitch := cmp_w - con_w) != 0:
        c_inst.add_ref(
            component=via_stack(
                x_range=(cmp.dxmin, cmp.dxmin + con_w),
                y_range=(cmp.dymin, cmp.dymax),
                base_layer=layer["comp"],
                metal_level=1,
            ),
            columns=2,
            column_pitch=column_pitch,
        )  # comp contact
    else:
        c_inst.add_ref(
            component=via_stack(
                x_range=(cmp.dxmin, cmp.dxmin + con_w),
                y_range=(cmp.dymin, cmp.dymax),
                base_layer=layer["comp"],
                metal_level=1,
            )
        )

    imp_rect = c_inst.add_ref(
        gf.components.rectangle(
            size=(
                cmp.dxsize + (2 * implant_enc[0]),
                cmp.dysize + (2 * implant_enc[1]),
            ),
            layer=implant_layer,
        )
    )
    imp_rect.dxmin = cmp.dxmin - implant_enc[0]
    imp_rect.dymin = cmp.dymin - implant_enc[1]

    poly = c_inst.add_ref(
        gf.components.rectangle(size=(lc, pl_l), layer=layer["poly2"])
    )

    poly.dxmin = cmp.dxmin + cmp_ext
    poly.dymin = cmp.dymin - pl_ext

    pl_con_el = via_stack(
        x_range=(poly.dxmin, poly.dxmax),
        y_range=(poly.dymin, poly.dymin + con_w),
        base_layer=layer["poly2"],
        metal_level=1,
    )

    if (column_pitch := pl_l - con_w) != 0:
        pl_con = c_inst.add_ref(component=pl_con_el, rows=2, row_pitch=pl_l - con_w)
    else:
        pl_con = c_inst.add_ref(component=pl_con_el)

    # Gate labels_generation

    if label == 1:
        c_inst.add_label(
            g_label,
            position=(
                pl_con.dxmin + (pl_con.dxsize / 2),
                pl_con.dymin + (pl_con_el.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

    pl_m1 = c_inst.add_ref(
        gf.components.rectangle(
            size=(pl_con.dxsize, pl_con.dysize), layer=layer["metal1"]
        )
    )
    pl_m1.dxmin = pl_con.dxmin
    pl_m1.dymin = pl_con.dymin

    return c_inst


@gf.cell
def cap_mos(
    type: str = "cap_nmos",
    lc: float = 0.1,
    wc: float = 0.1,
    volt: str = "3.3V",
    deepnwell: bool = False,
    pcmpgr: bool = False,
    label: bool = False,
    g_label: str = "",
    sd_label: str = "",
) -> gf.Component:
    """Usage:-
     used to draw NMOS capacitor (Outside DNWELL) by specifying parameters
    Arguments:-
     l      : Float of diff length
     w      : Float of diff width.
    """
    c = gf.Component("cap_mos_dev")

    con_size = 0.22
    con_sp = 0.28
    con_comp_enc = 0.07
    con_pl_enc = 0.07
    cmp_ext = 0.15 - con_comp_enc
    pl_ext = 0.17 - con_pl_enc

    np_enc_gate: float = 0.23
    np_enc_cmp: float = 0.16

    dg_enc_cmp = 0.24
    dg_enc_poly = 0.4
    lvpwell_enc_ncmp = 0.43
    dn_enc_lvpwell = 2.5

    grw = 0.36

    m1_w = 1
    pcmpgr_enc_dn = 2.5
    m1_ext = 0.82
    comp_pp_enc: float = 0.16
    dnwell_enc_pcmp = 1.1

    # end_cap: float = 0.22

    cmp_ed_w = con_size + (2 * con_comp_enc)
    cmp_w = (2 * (cmp_ed_w + cmp_ext)) + lc
    end_cap = pl_ext + cmp_ed_w

    pl_l = wc + (2 * end_cap)

    if "cap_nmos" in type:
        implant_layer = layer["nplus"]
    else:
        implant_layer = layer["pplus"]

    c_inst = c.add_ref(
        cap_mos_inst(
            cmp_w=cmp_w,
            lc=lc,
            wc=wc,
            pl_l=pl_l,
            cmp_ext=cmp_ed_w + cmp_ext,
            con_w=cmp_ed_w,
            pl_ext=end_cap,
            implant_layer=implant_layer,
            implant_enc=(np_enc_cmp, np_enc_gate),
            label=label,
            g_label=g_label,
        )
    )

    cmp_m1_polys = gf.Component(base=c_inst.cell.base).get_polygons_points()[
        layer["metal1"]
    ]
    cmp_m1_xmin = np.min(cmp_m1_polys[0][:, 0])
    cmp_m1_xmax = np.max(cmp_m1_polys[0][:, 0])
    cmp_m1_ymax = np.max(cmp_m1_polys[0][:, 1])

    # cmp_m1 = c.add_ref(gf.components.rectangle(size=(m1_w,w+m1_ext),layer=layer["metal1"]))
    cmp_m1_v = c.add_ref(
        component=gf.components.rectangle(
            size=(m1_w, wc + m1_ext), layer=layer["metal1"]
        ),
        rows=1,
        columns=2,
        column_pitch=m1_w + cmp_w - 2 * cmp_ed_w,
    )
    cmp_m1_v.dxmin = cmp_m1_xmin - (m1_w - (cmp_m1_xmax - cmp_m1_xmin))
    cmp_m1_v.dymax = cmp_m1_ymax

    cmp_m1_h = c.add_ref(
        gf.components.rectangle(size=(cmp_m1_v.dxsize, m1_w), layer=layer["metal1"])
    )
    cmp_m1_h.dxmin = cmp_m1_v.dxmin
    cmp_m1_h.dymax = cmp_m1_v.dymin

    # sd labels generation
    if label == 1:
        c.add_label(
            sd_label,
            position=(
                cmp_m1_h.dxmin + (cmp_m1_h.dxsize / 2),
                cmp_m1_h.dymin + (cmp_m1_h.dysize / 2),
            ),
            layer=layer["metal1_label"],
        )

    # dualgate

    if volt == "5/6V":
        dg = c.add_ref(
            gf.components.rectangle(
                size=(
                    c_inst.dxsize + (2 * dg_enc_cmp),
                    c_inst.dysize + (2 * dg_enc_poly),
                ),
                layer=layer["dualgate"],
            )
        )
        dg.dxmin = c_inst.dxmin - dg_enc_cmp
        dg.dymin = c_inst.dymin - dg_enc_poly

    cmp_polys = gf.Component(base=c_inst.cell.base).get_polygons_points()[layer["comp"]]
    cmp_xmin = np.min(cmp_polys[0][:, 0])
    cmp_ymin = np.min(cmp_polys[0][:, 1])
    cmp_xmax = np.max(cmp_polys[0][:, 0])
    cmp_ymax = np.max(cmp_polys[0][:, 1])

    if "_b" in type:
        if "cap_nmos" in type:
            nwell = c.add_ref(
                gf.components.rectangle(
                    size=(
                        cmp_xmax - cmp_xmin + (2 * np_enc_cmp),
                        cmp_ymax - cmp_ymin + (2 * np_enc_gate),
                    ),
                    layer=layer["nwell"],
                )
            )
            nwell.dxmin = cmp_xmin - np_enc_cmp
            nwell.dymin = cmp_ymin - np_enc_gate
        else:
            lvpwell = c.add_ref(
                gf.components.rectangle(
                    size=(
                        cmp_xmax - cmp_xmin + (2 * np_enc_cmp),
                        cmp_ymax - cmp_ymin + (2 * np_enc_gate),
                    ),
                    layer=layer["lvpwell"],
                )
            )

            lvpwell.dxmin = cmp_xmin - np_enc_cmp
            lvpwell.dymin = cmp_ymin - np_enc_gate

    if deepnwell == 1:
        if type == "cap_nmos":
            lvp_rect = c.add_ref(
                gf.components.rectangle(
                    size=(
                        c_inst.dxsize + (2 * lvpwell_enc_ncmp),
                        c_inst.dysize + (2 * lvpwell_enc_ncmp),
                    ),
                    layer=layer["lvpwell"],
                )
            )

            lvp_rect.dxmin = c_inst.dxmin - lvpwell_enc_ncmp
            lvp_rect.dymin = c_inst.dymin - lvpwell_enc_ncmp

            dn_rect = c.add_ref(
                gf.components.rectangle(
                    size=(
                        lvp_rect.dxsize + (2 * dn_enc_lvpwell),
                        lvp_rect.dysize + (2 * dn_enc_lvpwell),
                    ),
                    layer=layer["nwell"],
                )
            )

            dn_rect.dxmin = lvp_rect.dxmin - dn_enc_lvpwell
            dn_rect.dymin = lvp_rect.dymin - dn_enc_lvpwell

        else:
            dn_rect = c.add_ref(
                gf.components.rectangle(
                    size=(
                        c_inst.dxsize + (2 * dnwell_enc_pcmp),
                        c_inst.dysize + (2 * dnwell_enc_pcmp),
                    ),
                    layer=layer["nwell"],
                )
            )

            dn_rect.dxmin = c_inst.dxmin - dnwell_enc_pcmp
            dn_rect.dymin = c_inst.dymin - dnwell_enc_pcmp

        if pcmpgr == 1:
            c_temp_gr = gf.Component("temp_store guard ring")
            rect_pcmpgr_in = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (dn_rect.dxmax - dn_rect.dxmin) + 2 * pcmpgr_enc_dn,
                        (dn_rect.dymax - dn_rect.dymin) + 2 * pcmpgr_enc_dn,
                    ),
                    layer=layer["comp"],
                )
            )
            rect_pcmpgr_in.dmove(
                (dn_rect.dxmin - pcmpgr_enc_dn, dn_rect.dymin - pcmpgr_enc_dn)
            )
            rect_pcmpgr_out = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) + 2 * grw,
                        (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) + 2 * grw,
                    ),
                    layer=layer["comp"],
                )
            )
            rect_pcmpgr_out.dmove(
                (rect_pcmpgr_in.dxmin - grw, rect_pcmpgr_in.dymin - grw)
            )
            c.add_ref(
                gf.boolean(
                    A=rect_pcmpgr_out,
                    B=rect_pcmpgr_in,
                    operation="A-B",
                    layer=layer["comp"],
                )
            )  # guardring Bullk

            psdm_in = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (rect_pcmpgr_in.dxmax - rect_pcmpgr_in.dxmin) - 2 * comp_pp_enc,
                        (rect_pcmpgr_in.dymax - rect_pcmpgr_in.dymin) - 2 * comp_pp_enc,
                    ),
                    layer=layer["pplus"],
                )
            )
            psdm_in.dmove(
                (
                    rect_pcmpgr_in.dxmin + comp_pp_enc,
                    rect_pcmpgr_in.dymin + comp_pp_enc,
                )
            )
            psdm_out = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (rect_pcmpgr_out.dxmax - rect_pcmpgr_out.dxmin)
                        + 2 * comp_pp_enc,
                        (rect_pcmpgr_out.dymax - rect_pcmpgr_out.dymin)
                        + 2 * comp_pp_enc,
                    ),
                    layer=layer["pplus"],
                )
            )
            psdm_out.dmove(
                (
                    rect_pcmpgr_out.dxmin - comp_pp_enc,
                    rect_pcmpgr_out.dymin - comp_pp_enc,
                )
            )
            c.add_ref(
                gf.boolean(A=psdm_out, B=psdm_in, operation="A-B", layer=layer["pplus"])
            )  # psdm

            # generatingg contacts

            c.add_ref(
                via_generator(
                    x_range=(
                        rect_pcmpgr_in.dxmin + con_size,
                        rect_pcmpgr_in.dxmax - con_size,
                    ),
                    y_range=(rect_pcmpgr_out.dymin, rect_pcmpgr_in.dymin),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # bottom contact

            c.add_ref(
                via_generator(
                    x_range=(
                        rect_pcmpgr_in.dxmin + con_size,
                        rect_pcmpgr_in.dxmax - con_size,
                    ),
                    y_range=(rect_pcmpgr_in.dymax, rect_pcmpgr_out.dymax),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # upper contact

            c.add_ref(
                via_generator(
                    x_range=(rect_pcmpgr_out.dxmin, rect_pcmpgr_in.dxmin),
                    y_range=(
                        rect_pcmpgr_in.dymin + con_size,
                        rect_pcmpgr_in.dymax - con_size,
                    ),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # right contact

            c.add_ref(
                via_generator(
                    x_range=(rect_pcmpgr_in.dxmax, rect_pcmpgr_out.dxmax),
                    y_range=(
                        rect_pcmpgr_in.dymin + con_size,
                        rect_pcmpgr_in.dymax - con_size,
                    ),
                    via_enclosure=(con_comp_enc, con_comp_enc),
                    via_layer=layer["contact"],
                    via_size=(con_size, con_size),
                    via_spacing=(con_sp, con_sp),
                )
            )  # left contact

            comp_m1_in = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(rect_pcmpgr_in.dxsize, rect_pcmpgr_in.dysize),
                    layer=layer["metal1"],
                )
            )

            comp_m1_out = c_temp_gr.add_ref(
                gf.components.rectangle(
                    size=(
                        (comp_m1_in.dxsize) + 2 * grw,
                        (comp_m1_in.dysize) + 2 * grw,
                    ),
                    layer=layer["metal1"],
                )
            )
            comp_m1_out.dmove((rect_pcmpgr_in.dxmin - grw, rect_pcmpgr_in.dymin - grw))
            c.add_ref(
                gf.boolean(
                    A=rect_pcmpgr_out,
                    B=rect_pcmpgr_in,
                    operation="A-B",
                    layer=layer["metal1"],
                )
            )  # guardring metal1

    return c


if __name__ == "__main__":
    c = cap_mos()
    c.show()



from functools import partial
from math import ceil, floor

import gdsfactory as gf
from gdsfactory.typings import Float2, LayerSpec, Strs

from gf180mcu.guardring import pcmpgr_gen
from gf180mcu.layers import layer
from gf180mcu.via_generator import via_generator, via_stack

rectangle = partial(gf.components.rectangle, layer=layer["comp"])
rectangle_array = partial(gf.components.array, component=rectangle)


@gf.cell
def labels_gen(
    label_str: str = "",
    position: Float2 = (0.1, 0.1),
    layer: LayerSpec = layer["metal1_label"],
    label: bool = False,
    labels: Strs | None = None,
    label_valid_len: int = 1,
    index: int = 0,
) -> gf.Component:
    """Returns labels at given position when label is enabled.

    Args :
        label_str : string of the label.
        position : position of the label.
        layer : layer of the label.
        label : boolean of having the label.
        labels : list of given labels.
        label_valid_len : valid length of labels.
    """
    c = gf.Component()

    if label == 1:
        if len(labels) == label_valid_len:
            if label_str == "None":
                c.add_label(labels[index], position=position, layer=layer)
            else:
                c.add_label(label_str, position=position, layer=layer)

    return c


def get_patt_label(nl_b, nl, nt, nt_e, g_label, nl_u, nt_o):
    """Returns list of odd,even gate label patterns for alternating gate connection.

    Args:
        nl_b : number of bottom connected gates transistors.
        nl : number of transistor.
        nt : patterns of tansistor [with out redundancy].
        nt_e : number of transistor with even order.
        g_label : list of transistors gate label.
        nl_u :  number of upper connected gates transistors.
        nt_o : number of transistor with odd order.

    """
    g_label_e = []
    g_label_o = []

    if nt == len(g_label):
        for i in range(nl_b):
            for j in range(nl):
                if nt[j] == nt_e[i]:
                    g_label_e.extend(g_label[j])

        for i in range(nl_u):
            for j in range(nl):
                if nt[j] == nt_o[i]:
                    g_label_o.extend(g_label[j])

    return [g_label_e, g_label_o]


@gf.cell
def alter_interdig(
    sd_diff=rectangle,
    pc1=rectangle_array,
    pc2=rectangle_array,
    sd_l=0.36,
    nf=1,
    pat="",
    pc_x=0.1,
    pc_spacing=0.1,
    label: bool = False,
    g_label: Strs | None = None,
    nl: int = 1,
    patt_label: bool = False,
) -> gf.Component:
    """Returns interdigitation polygons of gate with alternating poly contacts.

    Args :
        sd_diff : source/drain diffusion rectangle.
        pc1 : first poly contact array.
        pc2 : second poly contact array.
        sd_l : source/drain length.
        nf : number of fingers.
        pat: string of the required pattern.
        poly_con : component of poly contact.
        sd_diff_inter : inter source/drain diffusion rectangle.
        l_gate : gate length.
        inter_sd_l : inter diffusion length.
        nf : number of fingers.
    """
    c_inst = gf.Component()
    sd_diff = gf.get_component(sd_diff)
    pc1 = gf.get_component(pc1)
    pc2 = gf.get_component(pc2)

    m2_spacing = 0.28
    via_size = (0.26, 0.26)
    via_enc = (0.06, 0.06)
    via_spacing = (0.26, 0.26)

    pat_o = []
    pat_e = []

    if pat:
        for i in range(int(nf)):
            if i % 2 == 0:
                pat_e.append(pat[i])
            else:
                pat_o.append(pat[i])

    nt = []
    [nt.append(dx) for dx in pat if dx not in nt]

    nt_o = []
    [nt_o.append(dx) for dx in pat_o if dx not in nt_o]

    nt_e = []
    [nt_e.append(dx) for dx in pat_e if dx not in nt_e]

    nl = len(nt)
    nl_b = len(nt_e)
    nl_u = len(nt_o)

    if pat:
        g_label_e, g_label_o = get_patt_label(nl_b, nl, nt, nt_e, g_label, nl_u, nt_o)

    m2_y = via_size[1] + 2 * via_enc[1]
    m2 = gf.components.rectangle(
        size=(sd_diff.dxmax - sd_diff.dxmin, m2_y),
        layer=layer["metal2"],
    )

    m2_arrb = c_inst.add_ref(
        component=m2,
        columns=1,
        rows=nl_b,
        row_pitch=-m2_y - m2_spacing,
    )
    m2_arrb.dmovey(pc1.dymin - m2_spacing - m2_y)

    m2_arru = c_inst.add_ref(
        component=m2,
        columns=1,
        rows=nl_u,
        row_pitch=m2_y + m2_spacing,
    )
    m2_arru.dmovey(pc2.dymax + m2_spacing)

    for i in range(nl_u):
        for j in range(floor(nf / 2)):
            if pat_o[j] == nt_o[i]:
                m1 = c_inst.add_ref(
                    gf.components.rectangle(
                        size=(
                            pc_x,
                            ((pc2.dymax + (i + 1) * (m2_spacing + m2_y)) - pc2.dymin),
                        ),
                        layer=layer["metal1"],
                    )
                )
                m1.dxmin = pc2.dxmin + j * (pc_spacing)
                m1.dymin = pc2.dymin

                via1_dr = via_generator(
                    x_range=(m1.dxmin, m1.dxmax),
                    y_range=(
                        m2_arru.dymin + i * (m2_y + m2_spacing),
                        m2_arru.dymin + i * (m2_y + m2_spacing) + m2_y,
                    ),
                    via_enclosure=via_enc,
                    via_layer=layer["via1"],
                    via_size=via_size,
                    via_spacing=via_spacing,
                )
                via1 = c_inst.add_ref(via1_dr)

                c_inst.add_ref(
                    labels_gen(
                        label_str="None",
                        position=(
                            (via1.dxmax + via1.dxmin) / 2,
                            (via1.dymax + via1.dymin) / 2,
                        ),
                        layer=layer["metal2_label"],
                        label=patt_label,
                        labels=pat_o,
                        label_valid_len=len(pat_o),
                        index=j,
                    )
                )

                # adding gate_label
                c_inst.add_ref(
                    labels_gen(
                        label_str="None",
                        position=(
                            m1.dxmin + (m1.dxsize / 2),
                            pc2.dymin + (pc2.dysize / 2),
                        ),
                        layer=layer["metal1_label"],
                        label=label,
                        labels=g_label_o,
                        label_valid_len=nl_u,
                        index=i,
                    )
                )

    for i in range(nl_b):
        for j in range(ceil(nf / 2)):
            if pat_e[j] == nt_e[i]:
                m1 = c_inst.add_ref(
                    gf.components.rectangle(
                        size=(
                            # poly_con.dxmax - poly_con.dxmin,
                            pc_x,
                            ((pc1.dymax + (i + 1) * (m2_spacing + m2_y)) - pc1.dymin),
                        ),
                        layer=layer["metal1"],
                    )
                )
                m1.dxmin = pc1.dxmin + j * (pc_spacing)
                m1.dymin = -(m1.dymax - m1.dymin) + (pc1.dymax)
                via1_dr = via_generator(
                    x_range=(m1.dxmin, m1.dxmax),
                    y_range=(
                        m2_arrb.dymax - i * (m2_spacing + m2_y) - m2_y,
                        m2_arrb.dymax - i * (m2_spacing + m2_y),
                    ),
                    via_enclosure=via_enc,
                    via_layer=layer["via1"],
                    via_size=via_size,
                    via_spacing=via_spacing,
                )
                via1 = c_inst.add_ref(via1_dr)

                c_inst.add_ref(
                    labels_gen(
                        label_str="None",
                        position=(
                            (via1.dxmax + via1.dxmin) / 2,
                            (via1.dymax + via1.dymin) / 2,
                        ),
                        layer=layer["metal2_label"],
                        label=patt_label,
                        labels=pat_e,
                        label_valid_len=len(pat_e),
                        index=j,
                    )
                )

                # adding gate_label
                c_inst.add_ref(
                    labels_gen(
                        label_str="None",
                        position=(
                            m1.dxmin + (m1.dxsize / 2),
                            pc1.dymin + (pc1.dysize / 2),
                        ),
                        layer=layer["metal1_label"],
                        label=label,
                        labels=g_label_e,
                        label_valid_len=nl_b,
                        index=i,
                    )
                )

    m3_x = via_size[0] + 2 * via_enc[0]
    m3_spacing = m2_spacing

    for i in range(nl_b):
        for j in range(nl_u):
            if nt_e[i] == nt_o[j]:
                m2_join_b = c_inst.add_ref(
                    gf.components.rectangle(
                        size=(
                            m2_y + sd_l + (i + 1) * (m3_spacing + m3_x),
                            m2_y,
                        ),
                        layer=layer["metal2"],
                    ).dmove(
                        (
                            m2_arrb.dxmin
                            - (m2_y + sd_l + (i + 1) * (m3_spacing + m3_x)),
                            m2_arrb.dymax - i * (m2_spacing + m2_y) - m2_y,
                        )
                    )
                )
                m2_join_u = c_inst.add_ref(
                    gf.components.rectangle(
                        size=(
                            m2_y + sd_l + (i + 1) * (m3_spacing + m3_x),
                            m2_y,
                        ),
                        layer=layer["metal2"],
                    ).dmove(
                        (
                            m2_arru.dxmin
                            - (m2_y + sd_l + (i + 1) * (m3_spacing + m3_x)),
                            m2_arru.dymin + j * (m2_spacing + m2_y),
                        )
                    )
                )
                m3 = c_inst.add_ref(
                    gf.components.rectangle(
                        size=(
                            m3_x,
                            m2_join_u.dymax - m2_join_b.dymin,
                        ),
                        layer=layer["metal1"],
                    )
                )
                m3.dmove((m2_join_b.dxmin, m2_join_b.dymin))
                via2_dr = via_generator(
                    x_range=(m3.dxmin, m3.dxmax),
                    y_range=(m2_join_b.dymin, m2_join_b.dymax),
                    via_enclosure=via_enc,
                    via_size=via_size,
                    via_layer=layer["via1"],
                    via_spacing=via_spacing,
                )
                c_inst.add_ref(
                    component=via2_dr,
                    columns=1,
                    rows=2,
                    row_pitch=m2_join_u.dymin - m2_join_b.dymin,
                )  # via2_draw
    return c_inst


@gf.cell
def interdigit(
    sd_diff=rectangle,
    pc1=rectangle_array,
    pc2=rectangle_array,
    poly_con=rectangle,
    sd_l: float = 0.15,
    nf=1,
    patt=[""],
    gate_con_pos="top",
    pc_x=0.1,
    pc_spacing=0.1,
    label: bool = False,
    g_label: Strs | None = None,
    patt_label: bool = False,
) -> gf.Component:
    """Returns interdigitation related polygons.

    Args :
        sd_diff : source/drain diffusion rectangle.
        pc1: first poly contact array.
        pc2: second poly contact array.
        poly_con: poly contact.
        sd_diff_inter : inter source/drain diffusion rectangle.
        l_gate : gate length.
        inter_sd_l : inter diffusion length.
        nf : number of fingers.
        pat : string of the required pattern.
        gate_con_pos : position of gate contact.
    """
    c_inst = gf.Component()

    sd_diff = gf.get_component(sd_diff)
    poly_con = gf.get_component(poly_con)
    pc1 = gf.get_component(pc1)
    pc2 = gf.get_component(pc2)

    if nf == len(patt):
        pat = list(patt)
        nt = []  # list to store the symbols of transistors and their number nt(number of transistors)
        [nt.append(dx) for dx in pat if dx not in nt]
        nl = len(nt)

        m2_spacing = 0.28
        via_size = (0.26, 0.26)
        via_enc = (0.06, 0.06)
        via_spacing = (0.26, 0.26)

        m2_y = via_size[1] + 2 * via_enc[1]
        m2 = gf.components.rectangle(
            size=(sd_diff.dxmax - sd_diff.dxmin, m2_y), layer=layer["metal2"]
        )

        if gate_con_pos == "alternating":
            c_inst.add_ref(
                alter_interdig(
                    sd_diff=sd_diff,
                    pc1=pc1,
                    pc2=pc2,
                    sd_l=sd_l,
                    nf=nf,
                    pat=pat,
                    pc_x=pc_x,
                    pc_spacing=pc_spacing,
                    label=label,
                    g_label=g_label,
                    nl=nl,
                    patt_label=patt_label,
                )
            )

        elif gate_con_pos == "top":
            m2_arr = c_inst.add_ref(
                component=m2,
                columns=1,
                rows=nl,
                row_pitch=m2.dymax - m2.dymin + m2_spacing,
            )
            m2_arr.dmovey(pc2.dymax + m2_spacing)

            for i in range(nl):
                for j in range(int(nf)):
                    if pat[j] == nt[i]:
                        m1 = c_inst.add_ref(
                            gf.components.rectangle(
                                size=(
                                    pc_x,
                                    # poly_con.dxmax - poly_con.dxmin,
                                    (
                                        (pc2.dymax + (i + 1) * (m2_spacing + m2_y))
                                        - ((1 - j % 2) * pc1.dymin)
                                        - (j % 2) * pc2.dymin
                                    ),
                                ),
                                layer=layer["metal1"],
                            )
                        )
                        m1.dxmin = pc1.dxmin + j * (pc2.dxmin - pc1.dxmin)
                        m1.dymin = pc1.dymin

                        via1_dr = via_generator(
                            x_range=(m1.dxmin, m1.dxmax),
                            y_range=(
                                m2_arr.dymin + i * (m2_spacing + m2_y),
                                m2_arr.dymin + i * (m2_spacing + m2_y) + m2_y,
                            ),
                            via_enclosure=via_enc,
                            via_layer=layer["via1"],
                            via_size=via_size,
                            via_spacing=via_spacing,
                        )
                        via1 = c_inst.add_ref(via1_dr)

                        c_inst.add_ref(
                            labels_gen(
                                label_str="None",
                                position=(
                                    (via1.dxmax + via1.dxmin) / 2,
                                    (via1.dymax + via1.dymin) / 2,
                                ),
                                layer=layer["metal2_label"],
                                label=patt_label,
                                labels=pat,
                                label_valid_len=nl,
                                index=j,
                            )
                        )

                        # adding gate_label
                        c_inst.add_ref(
                            labels_gen(
                                label_str="None",
                                position=(
                                    m1.dxmin + (m1.dxsize / 2),
                                    pc1.dymin + (pc1.dysize / 2),
                                ),
                                layer=layer["metal1_label"],
                                label=label,
                                labels=g_label,
                                label_valid_len=nl,
                                index=i,
                            )
                        )

        elif gate_con_pos == "bottom":
            m2_arr = c_inst.add_ref(
                component=m2,
                columns=1,
                rows=nl,
                row_pitch=-m2_y - m2_spacing,
            )
            m2_arr.dmovey(pc2.dymin - m2_spacing - m2_y)

            for i in range(nl):
                for j in range(int(nf)):
                    if pat[j] == nt[i]:
                        m1 = c_inst.add_ref(
                            gf.components.rectangle(
                                size=(
                                    # poly_con.dxmax - poly_con.dxmin,
                                    pc_x,
                                    (
                                        (pc1.dymax + (i + 1) * (m2_spacing + m2_y))
                                        - (j % 2) * pc1.dymin
                                        - (1 - j % 2) * pc2.dymin
                                    ),
                                ),
                                layer=layer["metal1"],
                            )
                        )
                        m1.dxmin = pc1.dxmin + j * (pc2.dxmin - pc1.dxmin)
                        m1.dymax = pc1.dymax

                        via1_dr = via_generator(
                            x_range=(m1.dxmin, m1.dxmax),
                            y_range=(
                                m2_arr.dymax - i * (m2_spacing + m2_y) - m2_y,
                                m2_arr.dymax - i * (m2_spacing + m2_y),
                            ),
                            via_enclosure=via_enc,
                            via_layer=layer["via1"],
                            via_size=via_size,
                            via_spacing=via_spacing,
                        )
                        via1 = c_inst.add_ref(via1_dr)

                        c_inst.add_ref(
                            labels_gen(
                                label_str="None",
                                position=(
                                    (via1.dxmax + via1.dxmin) / 2,
                                    (via1.dymax + via1.dymin) / 2,
                                ),
                                layer=layer["metal2_label"],
                                label=patt_label,
                                labels=pat,
                                label_valid_len=nl,
                                index=j,
                            )
                        )

                        # adding gate_label
                        c_inst.add_ref(
                            labels_gen(
                                label_str="None",
                                position=(
                                    m1.dxmin + (m1.dxsize / 2),
                                    pc1.dymin + (pc1.dysize / 2),
                                ),
                                layer=layer["metal1_label"],
                                label=label,
                                labels=g_label,
                                label_valid_len=nl,
                                index=i,
                            )
                        )

    return c_inst


def hv_gen(c, c_inst, volt, dg_encx: float = 0.1, dg_ency: float = 0.1) -> None:
    """Returns high voltage related polygons.

    Args :
        c_inst : dualgate enclosed component
        volt : operating voltage
        dg_encx : dualgate enclosure in x_direction
        dg_ency : dualgate enclosure in y_direction
    """
    if volt == "5V" or volt == "6V":
        dg = c.add_ref(
            gf.components.rectangle(
                size=(
                    c_inst.dxsize + (2 * dg_encx),
                    c_inst.dysize + (2 * dg_ency),
                ),
                layer=layer["dualgate"],
            )
        )
        dg.dxmin = c_inst.dxmin - dg_encx
        dg.dymin = c_inst.dymin - dg_ency

        if volt == "5V":
            v5x = c.add_ref(
                gf.components.rectangle(
                    size=(dg.dxsize, dg.dysize), layer=layer["v5_xtor"]
                )
            )
            v5x.dxmin = dg.dxmin
            v5x.dymin = dg.dymin

    # return c


def bulk_gr_gen(
    c: gf.Component,
    c_inst: gf.Component,
    comp_spacing: float = 0.1,
    poly2_comp_spacing: float = 0.1,
    volt: str = "3.3V",
    grw: float = 0.36,
    l_d: float = 0.1,
    implant_layer: LayerSpec = layer["pplus"],
    label: bool = False,
    sub_label: str = "",
    deepnwell: bool = False,
    pcmpgr: bool = False,
    nw_enc_pcmp: float = 0.1,
) -> None:
    """Returns guardring.

    Args :
        c_inst : component enclosed by guardring
        comp_spacing : spacing between comp polygons
        poly2_comp_spacing : spacing between comp and poly2 polygons
        volt : operating voltage
        grw : guardring width
        l_d : total diffusion length
        implant_layer : layer of comp implant (nplus,pplus)
    """
    # c = gf.Component()

    comp_pp_enc: float = 0.16

    con_size = 0.22
    con_sp = 0.28
    con_comp_enc = 0.07
    dg_enc_cmp = 0.24

    c_temp = gf.Component("temp_store")
    rect_bulk_in = c_temp.add_ref(
        gf.components.rectangle(
            size=(
                (c_inst.dxmax - c_inst.dxmin) + 2 * comp_spacing,
                (c_inst.dymax - c_inst.dymin) + 2 * poly2_comp_spacing,
            ),
            layer=layer["comp"],
        )
    )
    rect_bulk_in.dmove((c_inst.dxmin - comp_spacing, c_inst.dymin - poly2_comp_spacing))
    rect_bulk_out = c_temp.add_ref(
        gf.components.rectangle(
            size=(
                (rect_bulk_in.dxmax - rect_bulk_in.dxmin) + 2 * grw,
                (rect_bulk_in.dymax - rect_bulk_in.dymin) + 2 * grw,
            ),
            layer=layer["comp"],
        )
    )
    rect_bulk_out.dmove((rect_bulk_in.dxmin - grw, rect_bulk_in.dymin - grw))
    B = c.add_ref(
        gf.boolean(
            A=rect_bulk_out,
            B=rect_bulk_in,
            operation="A-B",
            layer=layer["comp"],
        )
    )

    psdm_in = c_temp.add_ref(
        gf.components.rectangle(
            size=(
                (rect_bulk_in.dxmax - rect_bulk_in.dxmin) - 2 * comp_pp_enc,
                (rect_bulk_in.dymax - rect_bulk_in.dymin) - 2 * comp_pp_enc,
            ),
            layer=layer["pplus"],
        )
    )
    psdm_in.dmove((rect_bulk_in.dxmin + comp_pp_enc, rect_bulk_in.dymin + comp_pp_enc))
    psdm_out = c_temp.add_ref(
        gf.components.rectangle(
            size=(
                (rect_bulk_out.dxmax - rect_bulk_out.dxmin) + 2 * comp_pp_enc,
                (rect_bulk_out.dymax - rect_bulk_out.dymin) + 2 * comp_pp_enc,
            ),
            layer=layer["pplus"],
        )
    )
    psdm_out.dmove(
        (
            rect_bulk_out.dxmin - comp_pp_enc,
            rect_bulk_out.dymin - comp_pp_enc,
        )
    )
    c.add_ref(
        gf.boolean(A=psdm_out, B=psdm_in, operation="A-B", layer=implant_layer)
    )  # implant_draw(pplus or nplus)

    # generatingg contacts

    c.add_ref(
        via_generator(
            x_range=(
                rect_bulk_in.dxmin + con_size,
                rect_bulk_in.dxmax - con_size,
            ),
            y_range=(rect_bulk_out.dymin, rect_bulk_in.dymin),
            via_enclosure=(con_comp_enc, con_comp_enc),
            via_layer=layer["contact"],
            via_size=(con_size, con_size),
            via_spacing=(con_sp, con_sp),
        )
    )  # bottom contact

    c.add_ref(
        via_generator(
            x_range=(
                rect_bulk_in.dxmin + con_size,
                rect_bulk_in.dxmax - con_size,
            ),
            y_range=(rect_bulk_in.dymax, rect_bulk_out.dymax),
            via_enclosure=(con_comp_enc, con_comp_enc),
            via_layer=layer["contact"],
            via_size=(con_size, con_size),
            via_spacing=(con_sp, con_sp),
        )
    )  # upper contact

    c.add_ref(
        via_generator(
            x_range=(rect_bulk_out.dxmin, rect_bulk_in.dxmin),
            y_range=(
                rect_bulk_in.dymin + con_size,
                rect_bulk_in.dymax - con_size,
            ),
            via_enclosure=(con_comp_enc, con_comp_enc),
            via_layer=layer["contact"],
            via_size=(con_size, con_size),
            via_spacing=(con_sp, con_sp),
        )
    )  # right contact

    c.add_ref(
        via_generator(
            x_range=(rect_bulk_in.dxmax, rect_bulk_out.dxmax),
            y_range=(
                rect_bulk_in.dymin + con_size,
                rect_bulk_in.dymax - con_size,
            ),
            via_enclosure=(con_comp_enc, con_comp_enc),
            via_layer=layer["contact"],
            via_size=(con_size, con_size),
            via_spacing=(con_sp, con_sp),
        )
    )  # left contact

    comp_m1_in = c_temp.add_ref(
        gf.components.rectangle(
            size=(
                (l_d) + 2 * comp_spacing,
                (c_inst.dymax - c_inst.dymin) + 2 * poly2_comp_spacing,
            ),
            layer=layer["metal1"],
        )
    )
    comp_m1_in.dmove((-comp_spacing, c_inst.dymin - poly2_comp_spacing))
    comp_m1_out = c_temp.add_ref(
        gf.components.rectangle(
            size=(
                (rect_bulk_in.dxmax - rect_bulk_in.dxmin) + 2 * grw,
                (rect_bulk_in.dymax - rect_bulk_in.dymin) + 2 * grw,
            ),
            layer=layer["metal1"],
        )
    )
    comp_m1_out.dmove((rect_bulk_in.dxmin - grw, rect_bulk_in.dymin - grw))
    c.add_ref(
        gf.boolean(
            A=rect_bulk_out,
            B=rect_bulk_in,
            operation="A-B",
            layer=layer["metal1"],
        )
    )  # metal1_guardring

    # c.add_ref(hv_gen(c_inst=B, volt=volt, dg_encx=dg_enc_cmp, dg_ency=dg_enc_cmp))
    hv_gen(c, c_inst=B, volt=volt, dg_encx=dg_enc_cmp, dg_ency=dg_enc_cmp)

    c.add_ref(
        labels_gen(
            label_str=sub_label,
            position=(
                B.dxmin + (grw + 2 * (comp_pp_enc)) / 2,
                B.dymin + (B.dysize / 2),
            ),
            layer=layer["metal1_label"],
            label=label,
            labels=[sub_label],
            label_valid_len=1,
        )
    )

    if implant_layer == layer["pplus"]:
        c.add_ref(
            nfet_deep_nwell(
                deepnwell=deepnwell,
                pcmpgr=pcmpgr,
                inst_size=(B.dxsize, B.dysize),
                inst_xmin=B.dxmin,
                inst_ymin=B.dymin,
                grw=grw,
            )
        )
    else:
        c.add_ref(
            pfet_deep_nwell(
                deepnwell=deepnwell,
                pcmpgr=pcmpgr,
                enc_size=(B.dxsize, B.dysize),
                enc_xmin=B.dxmin,
                enc_ymin=B.dymin,
                nw_enc_pcmp=nw_enc_pcmp,
                grw=grw,
            )
        )

    # return c


@gf.cell
def nfet_deep_nwell(
    deepnwell: bool = False,
    pcmpgr: bool = False,
    inst_size: Float2 = (0.1, 0.1),
    inst_xmin: float = 0.1,
    inst_ymin: float = 0.1,
    grw: float = 0.36,
) -> gf.Component:
    """Return nfet deepnwell.

    Args :
        deepnwell : boolean of having deepnwell
        pcmpgr : boolean of having deepnwell guardring
        inst_size : deepnwell enclosed size
        inst_xmin : deepnwell enclosed dxmin
        inst_ymin : deepnwell enclosed dymin
        grw : guardring width
    """
    c = gf.Component()

    dn_enc_lvpwell = 2.5
    lvpwell_enc_ncmp = 0.43

    if deepnwell == 1:
        lvp_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    inst_size[0] + (2 * lvpwell_enc_ncmp),
                    inst_size[1] + (2 * lvpwell_enc_ncmp),
                ),
                layer=layer["lvpwell"],
            )
        )

        lvp_rect.dxmin = inst_xmin - lvpwell_enc_ncmp
        lvp_rect.dymin = inst_ymin - lvpwell_enc_ncmp

        dn_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    lvp_rect.dxsize + (2 * dn_enc_lvpwell),
                    lvp_rect.dysize + (2 * dn_enc_lvpwell),
                ),
                layer=layer["dnwell"],
            )
        )

        dn_rect.dxmin = lvp_rect.dxmin - dn_enc_lvpwell
        dn_rect.dymin = lvp_rect.dymin - dn_enc_lvpwell

        if pcmpgr == 1:
            c.add_ref(pcmpgr_gen(dn_rect=dn_rect, grw=grw))

    return c


def add_inter_sd_labels(
    c: gf.Component,
    nf: int,
    sd_label: Strs | None,
    poly1: gf.Component,
    l_gate: float,
    inter_sd_l: float,
    sd_diff_intr: gf.Component,
    label: bool,
    layer: LayerSpec,
    con_bet_fin: int,
) -> None:
    """Adds label to intermediate source/drain diffusion.

    Args :
        c : instance component of the device
        nf : number of fingers
        sd_label : required source and drain labels list
        poly1 : component of poly array
        l_gate : length of fet gate
        inter_sd_l : length of intermediate source/drain diffusion
        sd_diff_inter : component of intermediate source/drain polygon
        label: boolean of having labels
        layer : layer of label
        con_bet_fin : boolean of having contact between fingers
    """
    if con_bet_fin == 1:
        label_layer = layer["metal1_label"]
    else:
        label_layer = layer["comp_label"]

    for i in range(int(nf - 1)):
        c.add_ref(
            labels_gen(
                label_str="None",
                position=(
                    poly1.dxmin + l_gate + (inter_sd_l / 2) + i * (l_gate + inter_sd_l),
                    sd_diff_intr.dymin + (sd_diff_intr.dysize / 2),
                ),
                layer=label_layer,
                label=label,
                labels=sd_label,
                label_valid_len=nf + 1,
                index=i + 1,
            )
        )


def add_gate_labels(
    c: gf.Component,
    g_label: Strs | None,
    pc1: gf.Component,
    c_pc: gf.Component,
    pc_spacing: float,
    nc1: int,
    nc2: int,
    pc2: gf.Component,
    label: bool,
    layer: LayerSpec,
    nf: int,
) -> None:
    """Adds gate label when label is enabled.

    Args :
        c : instance component of the device
        g_label : required gate labels list
        pc1 : component of poly array1
        c_pc : component of poly array element
        pc_spacing : float of space between labels
        nc1 : number of columns in poly array1
        nc2 : number of columns in poly array2
        pc2 : component of poly array2
        label : boolean of having labels
        layer : layer of labels
        nf : number of fingers
    """
    for i in range(nc1):
        c.add_ref(
            labels_gen(
                label_str="None",
                position=(
                    pc1.dxmin + (c_pc.dxsize / 2) + i * (pc_spacing),
                    pc1.dymin + (c_pc.dysize / 2),
                ),
                layer=layer["metal1_label"],
                label=label,
                labels=g_label,
                label_valid_len=nf,
                index=2 * i,
            )
        )

    for i in range(nc2):
        c.add_ref(
            labels_gen(
                label_str="None",
                position=(
                    pc2.dxmin + (c_pc.dxsize / 2) + i * (pc_spacing),
                    pc2.dymin + (c_pc.dysize / 2),
                ),
                layer=layer["metal1_label"],
                label=label,
                labels=g_label,
                label_valid_len=nf,
                index=(2 * i) + 1,
            )
        )


@gf.cell
def nfet(
    l_gate: float = 0.28,
    w_gate: float = 0.22,
    sd_con_col: int = 1,
    inter_sd_l: float = 0.24,
    nf: int = 1,
    grw: float = 0.22,
    volt: str = "3.3V",
    bulk: str = "None",
    con_bet_fin: int = 1,
    gate_con_pos: str = "alternating",
    interdig: int = 0,
    patt: str = "",
    deepnwell: int = 0,
    pcmpgr: int = 0,
    label: bool = False,
    sd_label: Strs | None = [],
    g_label: Strs = (),
    sub_label: str = "",
    patt_label: bool = False,
) -> gf.Component:
    """Return nfet.

    Args:
        l : Float of gate length
        w : Float of gate width
        sd_l : Float of source and drain diffusion length
        inter_sd_l : Float of source and drain diffusion length between fingers
        nf : integer of number of fingers
        M : integer of number of multipliers
        grw : guard ring width when enabled
        type : string of the device type
        bulk : String of bulk connection type (None, Bulk Tie, Guard Ring)
        con_bet_fin : boolean of having contacts for diffusion between fingers
        gate_con_pos : string of choosing the gate contact position (bottom, top, alternating )

    """
    # used layers and dimensions

    end_cap: float = 0.22
    if volt == "3.3V":
        comp_spacing: float = 0.28
    else:
        comp_spacing: float = 0.36

    gate_np_enc: float = 0.23
    comp_np_enc: float = 0.16
    comp_pp_enc: float = 0.16
    poly2_spacing: float = 0.24
    pc_ext: float = 0.04

    con_size = 0.22
    con_sp = 0.28
    con_comp_enc = 0.07
    con_pl_enc = 0.07
    dg_enc_cmp = 0.24
    dg_enc_poly = 0.4

    sd_l_con = (
        ((sd_con_col) * con_size) + ((sd_con_col - 1) * con_sp) + 2 * con_comp_enc
    )
    sd_l = sd_l_con

    # gds components to store a single instance and the generated device
    c = gf.Component()

    c_inst = gf.Component()

    # generatingg sd diffusion

    if interdig == 1 and nf > 1 and nf != len(patt) and patt != "":
        nf = len(patt)

    l_d = (
        nf * l_gate + (nf - 1) * inter_sd_l + 2 * (con_comp_enc)
    )  # diffution total length
    rect_d_intr = gf.components.rectangle(size=(l_d, w_gate), layer=layer["comp"])
    sd_diff_intr = c_inst.add_ref(rect_d_intr)

    #  generatingg sd contacts

    if w_gate <= con_size + 2 * con_comp_enc:
        cmpc_y = con_comp_enc + con_size + con_comp_enc

    else:
        cmpc_y = w_gate

    cmpc_size = (sd_l_con, cmpc_y)

    sd_diff = c_inst.add_ref(
        component=gf.components.rectangle(size=cmpc_size, layer=layer["comp"]),
        rows=1,
        columns=2,
        column_pitch=(cmpc_size[0] + sd_diff_intr.dxsize),
    )

    sd_diff.dxmin = sd_diff_intr.dxmin - cmpc_size[0]
    sd_diff.dymin = sd_diff_intr.dymin - (sd_diff.dysize - sd_diff_intr.dysize) / 2

    sd_con = via_stack(
        x_range=(sd_diff.dxmin, sd_diff_intr.dxmin),
        y_range=(sd_diff.dymin, sd_diff.dymax),
        base_layer=layer["comp"],
        metal_level=1,
    )
    if (
        column_pitch := sd_l + nf * l_gate + (nf - 1) * inter_sd_l + 2 * (con_comp_enc)
    ) != 0:
        c_inst.add_ref(component=sd_con, columns=2, column_pitch=column_pitch)
    else:
        c_inst.add_ref(component=sd_con)

    if con_bet_fin == 1 and nf > 1:
        inter_sd_con = via_stack(
            x_range=(
                sd_diff_intr.dxmin + con_comp_enc + l_gate,
                sd_diff_intr.dxmin + con_comp_enc + l_gate + inter_sd_l,
            ),
            y_range=(0, w_gate),
            base_layer=layer["comp"],
            metal_level=1,
        )
        c_inst.add_ref(
            component=inter_sd_con,
            columns=nf - 1,
            rows=1,
            column_pitch=(l_gate + inter_sd_l),
        )

    ### adding source/drain labels
    c.add_ref(
        labels_gen(
            label_str="None",
            position=(
                sd_diff.dxmin + (sd_l / 2),
                sd_diff.dymin + (sd_diff.dysize / 2),
            ),
            layer=layer["metal1_label"],
            label=label,
            labels=sd_label,
            label_valid_len=nf + 1,
            index=0,
        )
    )

    c.add_ref(
        labels_gen(
            label_str="None",
            position=(
                sd_diff.dxmax - (sd_l / 2),
                sd_diff.dymin + (sd_diff.dysize / 2),
            ),
            layer=layer["metal1_label"],
            label=label,
            labels=sd_label,
            label_valid_len=nf + 1,
            index=nf,
        )
    )

    # generatingg poly

    if l_gate <= con_size + 2 * con_pl_enc:
        pc_x = con_pl_enc + con_size + con_pl_enc

    else:
        pc_x = l_gate

    pc_size = (pc_x, con_pl_enc + con_size + con_pl_enc)

    c_pc = gf.Component()

    rect_pc = c_pc.add_ref(gf.components.rectangle(size=pc_size, layer=layer["poly2"]))

    poly_con = via_stack(
        x_range=(rect_pc.dxmin, rect_pc.dxmax),
        y_range=(rect_pc.dymin, rect_pc.dymax),
        base_layer=layer["poly2"],
        metal_level=1,
        li_enc_dir="H",
    )
    c_pc.add_ref(poly_con)

    if nf == 1:
        poly = c_inst.add_ref(
            gf.components.rectangle(
                size=(l_gate, w_gate + 2 * end_cap), layer=layer["poly2"]
            )
        )
        poly.dxmin = sd_diff_intr.dxmin + con_comp_enc
        poly.dymin = sd_diff_intr.dymin - end_cap

        if gate_con_pos == "bottom":
            mv = 0
            nr = 1
        elif gate_con_pos == "top":
            mv = pc_size[1] + w_gate + 2 * end_cap
            nr = 1
        else:
            mv = 0
            nr = 2

        pc = c_inst.add_ref(
            component=c_pc, rows=nr, row_pitch=pc_size[1] + w_gate + 2 * end_cap
        )
        pc.dmove((poly.dxmin - ((pc_x - l_gate) / 2), -pc_size[1] - end_cap + mv))

        # gate_lablel
        c.add_ref(
            labels_gen(
                label_str="None",
                position=(pc.dxmin + c_pc.dxsize / 2, pc.dymin + c_pc.dysize / 2),
                layer=layer["metal1_label"],
                label=label,
                labels=g_label,
                label_valid_len=nf,
                index=0,
            )
        )

    else:
        w_p1 = end_cap + w_gate + end_cap  # poly total width

        if inter_sd_l < (poly2_spacing + 2 * pc_ext):
            if gate_con_pos == "alternating":
                w_p1 += 0.2
                w_p2 = w_p1
                e_c = 0.2
            else:
                w_p2 = w_p1 + con_pl_enc + con_size + con_pl_enc + poly2_spacing + 0.1
                e_c = 0

            if gate_con_pos == "bottom":
                p_mv = -end_cap - (w_p2 - w_p1)
            else:
                p_mv = -end_cap

        else:
            w_p2 = w_p1
            p_mv = -end_cap
            e_c = 0

        rect_p1 = gf.components.rectangle(size=(l_gate, w_p1), layer=layer["poly2"])
        rect_p2 = gf.components.rectangle(size=(l_gate, w_p2), layer=layer["poly2"])
        poly1 = c_inst.add_ref(
            rect_p1,
            rows=1,
            columns=ceil(nf / 2),
            column_pitch=2 * (inter_sd_l + l_gate),
        )
        poly1.dxmin = sd_diff_intr.dxmin + con_comp_enc
        poly1.dymin = sd_diff_intr.dymin - end_cap - e_c

        poly2 = c_inst.add_ref(
            rect_p2,
            rows=1,
            columns=floor(nf / 2),
            column_pitch=2 * (inter_sd_l + l_gate),
        )
        poly2.dxmin = poly1.dxmin + l_gate + inter_sd_l
        poly2.dymin = p_mv

        # generatingg poly contacts setups

        if gate_con_pos == "bottom":
            mv_1 = 0
            mv_2 = -(w_p2 - w_p1)
        elif gate_con_pos == "top":
            mv_1 = pc_size[1] + w_p1
            mv_2 = pc_size[1] + w_p2
        else:
            mv_1 = -e_c
            mv_2 = pc_size[1] + w_p2

        nc1 = ceil(nf / 2)
        nc2 = floor(nf / 2)

        pc_spacing = 2 * (inter_sd_l + l_gate)

        # generatingg poly contacts

        pc1 = c_inst.add_ref(
            component=c_pc, rows=1, columns=nc1, column_pitch=pc_spacing
        )
        pc1.dmove((poly1.dxmin - ((pc_x - l_gate) / 2), -pc_size[1] - end_cap + mv_1))

        pc2 = c_inst.add_ref(
            component=c_pc, rows=1, columns=nc2, column_pitch=pc_spacing
        )
        pc2.dmove(
            (
                poly1.dxmin - ((pc_x - l_gate) / 2) + (inter_sd_l + l_gate),
                -pc_size[1] - end_cap + mv_2,
            )
        )

        add_inter_sd_labels(
            c,
            nf,
            sd_label,
            poly1,
            l_gate,
            inter_sd_l,
            sd_diff_intr,
            label,
            layer,
            con_bet_fin,
        )

        # add_gate_labels(c, g_label, pc1, c_pc, pc_spacing, nc1, nc2, pc2, label, layer, nf)

        if interdig == 1:
            c_inst.add_ref(
                interdigit(
                    sd_diff=sd_diff,
                    pc1=pc1,
                    pc2=pc2,
                    poly_con=poly_con,
                    sd_l=sd_l,
                    nf=nf,
                    patt=patt,
                    gate_con_pos=gate_con_pos,
                    pc_x=pc_x,
                    pc_spacing=pc_spacing,
                    label=label,
                    g_label=g_label,
                    patt_label=patt_label,
                )
            )
        else:
            add_gate_labels(
                c, g_label, pc1, c_pc, pc_spacing, nc1, nc2, pc2, label, layer, nf
            )

    # generatingg bulk
    if bulk == "None":
        nplus = c_inst.add_ref(
            gf.components.rectangle(
                size=(sd_diff.dxsize + 2 * comp_np_enc, w_gate + 2 * gate_np_enc),
                layer=layer["nplus"],
            )
        )
        nplus.dxmin = sd_diff.dxmin - comp_np_enc
        nplus.dymin = sd_diff_intr.dymin - gate_np_enc

    elif bulk == "Bulk Tie":
        rect_bulk = c_inst.add_ref(
            gf.components.rectangle(
                size=(sd_l + con_sp, sd_diff.dysize), layer=layer["comp"]
            )
        )
        rect_bulk.dxmin = sd_diff.dxmax
        rect_bulk.dymin = sd_diff.dymin
        nsdm = c_inst.add_ref(
            gf.components.rectangle(
                size=(
                    sd_diff.dxmax - sd_diff.dxmin + comp_np_enc,
                    w_gate + 2 * gate_np_enc,
                ),
                layer=layer["nplus"],
            )
        )
        nsdm.dxmin = sd_diff.dxmin - comp_np_enc
        nsdm.dymin = sd_diff_intr.dymin - gate_np_enc
        psdm = c_inst.add_ref(
            gf.components.rectangle(
                size=(
                    rect_bulk.dxmax - rect_bulk.dxmin + comp_pp_enc,
                    w_gate + 2 * comp_pp_enc,
                ),
                layer=layer["pplus"],
            )
        )
        psdm.connect("e1", nsdm.ports["e3"])

        bulk_con = via_stack(
            x_range=(rect_bulk.dxmin + 0.1, rect_bulk.dxmax - 0.1),
            y_range=(rect_bulk.dymin, rect_bulk.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
        c_inst.add_ref(bulk_con)

        c.add_ref(
            labels_gen(
                label_str=sub_label,
                position=(
                    bulk_con.dxmin + bulk_con.dxsize / 2,
                    bulk_con.dymin + bulk_con.dysize / 2,
                ),
                layer=layer["metal1_label"],
                label=label,
                labels=[sub_label],
                label_valid_len=1,
            )
        )

    if bulk == "Guard Ring":
        nsdm = c_inst.add_ref(
            gf.components.rectangle(
                size=(sd_diff.dxsize + 2 * comp_np_enc, w_gate + 2 * gate_np_enc),
                layer=layer["nplus"],
            )
        )
        nsdm.dxmin = sd_diff.dxmin - comp_np_enc
        nsdm.dymin = sd_diff_intr.dymin - gate_np_enc
        c.add_ref(c_inst)

        # b_gr = c.add_ref(
        bulk_gr_gen(
            c,
            c_inst=c_inst,
            comp_spacing=comp_spacing,
            poly2_comp_spacing=comp_spacing,
            volt=volt,
            grw=grw,
            l_d=l_d,
            implant_layer=layer["pplus"],
            label=label,
            sub_label=sub_label,
            deepnwell=deepnwell,
            pcmpgr=pcmpgr,
        )

    # if bulk != "Guard Ring":
    else:
        c.add_ref(c_inst)

        inst_size = (c_inst.dxsize, c_inst.dysize)
        inst_xmin = c_inst.dxmin
        inst_ymin = c_inst.dymin

        # c.add_ref(
        #     hv_gen(c_inst=c_inst, volt=volt, dg_encx=dg_enc_cmp, dg_ency=dg_enc_poly)
        # )
        hv_gen(c, c_inst=c_inst, volt=volt, dg_encx=dg_enc_cmp, dg_ency=dg_enc_poly)

        c.add_ref(
            nfet_deep_nwell(
                deepnwell=deepnwell,
                pcmpgr=pcmpgr,
                inst_size=inst_size,
                inst_xmin=inst_xmin,
                inst_ymin=inst_ymin,
                grw=grw,
            )
        )

    return c


@gf.cell
def pfet_deep_nwell(
    deepnwell: bool = False,
    pcmpgr: bool = False,
    enc_size: Float2 = (0.1, 0.1),
    enc_xmin: float = 0.1,
    enc_ymin: float = 0.1,
    nw_enc_pcmp: float = 0.1,
    grw: float = 0.36,
) -> gf.Component:
    """Returns pfet well related polygons.

    Args :
        deepnwell : boolaen of having deepnwell
        pcmpgr : boolean of having deepnwell guardring
        enc_size : enclosed size
        enc_xmin : enclosed dxmin
        enc_ymin : enclosed dymin
        nw_enc_pcmp : nwell enclosure of pcomp
        grw : guardring width
    """
    c = gf.Component()

    dnwell_enc_pcmp = 1.1

    if deepnwell == 1:
        dn_rect = c.add_ref(
            gf.components.rectangle(
                size=(
                    enc_size[0] + (2 * dnwell_enc_pcmp),
                    enc_size[1] + (2 * dnwell_enc_pcmp),
                ),
                layer=layer["dnwell"],
            )
        )

        dn_rect.dxmin = enc_xmin - dnwell_enc_pcmp
        dn_rect.dymin = enc_ymin - dnwell_enc_pcmp

        if pcmpgr == 1:
            c.add_ref(pcmpgr_gen(dn_rect=dn_rect, grw=grw))

    else:
        # nwell generation
        nw = c.add_ref(
            gf.components.rectangle(
                size=(
                    enc_size[0] + (2 * nw_enc_pcmp),
                    enc_size[1] + (2 * nw_enc_pcmp),
                ),
                layer=layer["nwell"],
            )
        )
        nw.dxmin = enc_xmin - nw_enc_pcmp
        nw.dymin = enc_ymin - nw_enc_pcmp

    return c


@gf.cell
def pfet(
    l_gate: float = 0.28,
    w_gate: float = 0.22,
    sd_con_col: int = 1,
    inter_sd_l: float = 0.24,
    nf: int = 1,
    grw: float = 0.22,
    volt: str = "3.3V",
    bulk: str = "None",
    con_bet_fin: int = 1,
    gate_con_pos: str = "alternating",
    interdig: int = 0,
    patt: str = "",
    deepnwell: int = 0,
    pcmpgr: int = 0,
    label: bool = False,
    sd_label: Strs | None = (),
    g_label: Strs = (),
    sub_label: str = "",
    patt_label: bool = False,
) -> gf.Component:
    """Return pfet.

    Args:
        l : Float of gate length
        w : Float of gate width
        sd_l : Float of source and drain diffusion length
        inter_sd_l : Float of source and drain diffusion length between fingers
        nf : integer of number of fingers
        M : integer of number of multipliers
        grw : guard ring width when enabled
        type : string of the device type
        bulk : String of bulk connection type (None, Bulk Tie, Guard Ring)
        con_bet_fin : boolean of having contacts for diffusion between fingers
        gate_con_pos : string of choosing the gate contact position (bottom, top, alternating )

    """
    # used layers and dimensions

    end_cap: float = 0.22
    if volt == "3.3V":
        comp_spacing: float = 0.28
        nw_enc_pcmp = 0.43
    else:
        comp_spacing: float = 0.36
        nw_enc_pcmp = 0.6

    gate_pp_enc: float = 0.23
    comp_np_enc: float = 0.16
    comp_pp_enc: float = 0.16
    poly2_spacing: float = 0.24
    pc_ext: float = 0.04

    con_size = 0.22
    con_sp = 0.28
    con_comp_enc = 0.07
    con_pl_enc = 0.07
    dg_enc_cmp = 0.24
    dg_enc_poly = 0.4

    sd_l_con = (
        ((sd_con_col) * con_size) + ((sd_con_col - 1) * con_sp) + 2 * con_comp_enc
    )
    sd_l = sd_l_con

    # gds components to store a single instance and the generated device
    c = gf.Component()

    c_inst = gf.Component()

    # generatingg sd diffusion

    if interdig == 1 and nf > 1 and nf != len(patt) and patt != "":
        nf = len(patt)

    l_d = (
        nf * l_gate + (nf - 1) * inter_sd_l + 2 * (con_comp_enc)
    )  # diffution total length
    rect_d_intr = gf.components.rectangle(size=(l_d, w_gate), layer=layer["comp"])
    sd_diff_intr = c_inst.add_ref(rect_d_intr)

    # generatingg sd contacts

    if w_gate <= con_size + 2 * con_comp_enc:
        cmpc_y = con_comp_enc + con_size + con_comp_enc

    else:
        cmpc_y = w_gate

    cmpc_size = (sd_l_con, cmpc_y)

    sd_diff = c_inst.add_ref(
        component=gf.components.rectangle(size=cmpc_size, layer=layer["comp"]),
        rows=1,
        columns=2,
        column_pitch=(cmpc_size[0] + sd_diff_intr.dxsize),
    )
    sd_diff.dxmin = sd_diff_intr.dxmin - cmpc_size[0]
    sd_diff.dymin = sd_diff_intr.dymin - (sd_diff.dysize - sd_diff_intr.dysize) / 2

    sd_con = via_stack(
        x_range=(sd_diff.dxmin, sd_diff_intr.dxmin),
        y_range=(sd_diff.dymin, sd_diff.dymax),
        base_layer=layer["comp"],
        metal_level=1,
    )
    c_inst.add_ref(
        component=sd_con,
        columns=2,
        rows=1,
        column_pitch=sd_l + nf * l_gate + (nf - 1) * inter_sd_l + 2 * (con_comp_enc),
    )

    if con_bet_fin == 1 and nf > 1:
        inter_sd_con = via_stack(
            x_range=(
                sd_diff_intr.dxmin + con_comp_enc + l_gate,
                sd_diff_intr.dxmin + con_comp_enc + l_gate + inter_sd_l,
            ),
            y_range=(0, w_gate),
            base_layer=layer["comp"],
            metal_level=1,
        )
        c_inst.add_ref(
            component=inter_sd_con,
            columns=nf - 1,
            rows=1,
            column_pitch=l_gate + inter_sd_l,
        )

    ### adding source/drain labels
    c.add_ref(
        labels_gen(
            label_str="None",
            position=(
                sd_diff.dxmin + (sd_l / 2),
                sd_diff.dymin + (sd_diff.dysize / 2),
            ),
            layer=layer["metal1_label"],
            label=label,
            labels=sd_label,
            label_valid_len=nf + 1,
            index=0,
        )
    )

    c.add_ref(
        labels_gen(
            label_str="None",
            position=(
                sd_diff.dxmax - (sd_l / 2),
                sd_diff.dymin + (sd_diff.dysize / 2),
            ),
            layer=layer["metal1_label"],
            label=label,
            labels=sd_label,
            label_valid_len=nf + 1,
            index=nf,
        )
    )

    # generatingg poly

    if l_gate <= con_size + 2 * con_pl_enc:
        pc_x = con_pl_enc + con_size + con_pl_enc

    else:
        pc_x = l_gate

    pc_size = (pc_x, con_pl_enc + con_size + con_pl_enc)

    c_pc = gf.Component()

    rect_pc = c_pc.add_ref(gf.components.rectangle(size=pc_size, layer=layer["poly2"]))

    poly_con = via_stack(
        x_range=(rect_pc.dxmin, rect_pc.dxmax),
        y_range=(rect_pc.dymin, rect_pc.dymax),
        base_layer=layer["poly2"],
        metal_level=1,
        li_enc_dir="H",
    )
    c_pc.add_ref(poly_con)

    if nf == 1:
        poly = c_inst.add_ref(
            gf.components.rectangle(
                size=(l_gate, w_gate + 2 * end_cap), layer=layer["poly2"]
            )
        )
        poly.dxmin = sd_diff_intr.dxmin + con_comp_enc
        poly.dymin = sd_diff_intr.dymin - end_cap

        if gate_con_pos == "bottom":
            mv = 0
            nr = 1
        elif gate_con_pos == "top":
            mv = pc_size[1] + w_gate + 2 * end_cap
            nr = 1
        else:
            mv = 0
            nr = 2

        pc = c_inst.add_ref(
            component=c_pc,
            rows=nr,
            columns=1,
            row_pitch=pc_size[1] + w_gate + 2 * end_cap,
        )
        pc.dmove((poly.dxmin - ((pc_x - l_gate) / 2), -pc_size[1] - end_cap + mv))

        # gate_lablel
        c.add_ref(
            labels_gen(
                label_str="None",
                position=(pc.dxmin + c_pc.dxsize / 2, pc.dymin + c_pc.dysize / 2),
                layer=layer["metal1_label"],
                label=label,
                labels=g_label,
                label_valid_len=nf,
                index=0,
            )
        )

    else:
        w_p1 = end_cap + w_gate + end_cap  # poly total width

        if inter_sd_l < (poly2_spacing + 2 * pc_ext):
            if gate_con_pos == "alternating":
                w_p1 += 0.2
                w_p2 = w_p1
                e_c = 0.2
            else:
                w_p2 = w_p1 + con_pl_enc + con_size + con_pl_enc + poly2_spacing + 0.1
                e_c = 0

            if gate_con_pos == "bottom":
                p_mv = -end_cap - (w_p2 - w_p1)
            else:
                p_mv = -end_cap

        else:
            w_p2 = w_p1
            p_mv = -end_cap
            e_c = 0

        rect_p1 = gf.components.rectangle(size=(l_gate, w_p1), layer=layer["poly2"])
        rect_p2 = gf.components.rectangle(size=(l_gate, w_p2), layer=layer["poly2"])
        poly1 = c_inst.add_ref(
            rect_p1,
            rows=1,
            columns=ceil(nf / 2),
            column_pitch=2 * (inter_sd_l + l_gate),
        )
        poly1.dxmin = sd_diff_intr.dxmin + con_comp_enc

        poly2 = c_inst.add_ref(
            rect_p2,
            rows=1,
            columns=floor(nf / 2),
            column_pitch=2 * (inter_sd_l + l_gate),
        )
        poly2.dxmin = poly1.dxmin + l_gate + inter_sd_l
        poly2.dymin = p_mv

        # generatingg poly contacts setups

        if gate_con_pos == "bottom":
            mv_1 = 0
            mv_2 = -(w_p2 - w_p1)
        elif gate_con_pos == "top":
            mv_1 = pc_size[1] + w_p1
            mv_2 = pc_size[1] + w_p2
        else:
            mv_1 = -e_c
            mv_2 = pc_size[1] + w_p2

        nc1 = ceil(nf / 2)
        nc2 = floor(nf / 2)

        pc_spacing = 2 * (inter_sd_l + l_gate)

        # generatingg poly contacts

        pc1 = c_inst.add_ref(
            component=c_pc, rows=1, columns=nc1, column_pitch=pc_spacing
        )
        pc1.dmove((poly1.dxmin - ((pc_x - l_gate) / 2), -pc_size[1] - end_cap + mv_1))

        pc2 = c_inst.add_ref(
            component=c_pc, rows=1, columns=nc2, column_pitch=pc_spacing
        )
        pc2.dmove(
            (
                poly1.dxmin - ((pc_x - l_gate) / 2) + (inter_sd_l + l_gate),
                -pc_size[1] - end_cap + mv_2,
            )
        )

        add_inter_sd_labels(
            c,
            nf,
            sd_label,
            poly1,
            l_gate,
            inter_sd_l,
            sd_diff_intr,
            label,
            layer,
            con_bet_fin,
        )

        add_gate_labels(
            c, g_label, pc1, c_pc, pc_spacing, nc1, nc2, pc2, label, layer, nf
        )

        if interdig == 1:
            c_inst.add_ref(
                interdigit(
                    sd_diff=sd_diff,
                    pc1=pc1,
                    pc2=pc2,
                    poly_con=poly_con,
                    sd_l=sd_l,
                    nf=nf,
                    patt=patt,
                    gate_con_pos=gate_con_pos,
                    pc_x=pc_x,
                    pc_spacing=pc_spacing,
                    label=label,
                    g_label=g_label,
                    patt_label=patt_label,
                )
            )

    # generatingg bulk
    if bulk == "None":
        pplus = c_inst.add_ref(
            gf.components.rectangle(
                size=(sd_diff.dxsize + 2 * comp_pp_enc, w_gate + 2 * gate_pp_enc),
                layer=layer["pplus"],
            )
        )
        pplus.dxmin = sd_diff.dxmin - comp_pp_enc
        pplus.dymin = sd_diff_intr.dymin - gate_pp_enc

        c.add_ref(c_inst)

        # deep nwell and nwell generation

        c.add_ref(
            pfet_deep_nwell(
                deepnwell=deepnwell,
                pcmpgr=pcmpgr,
                enc_size=(sd_diff.dxsize, sd_diff.dysize),
                enc_xmin=sd_diff.dxmin,
                enc_ymin=sd_diff.dymin,
                nw_enc_pcmp=nw_enc_pcmp,
                grw=grw,
            )
        )

        # dualgate generation

        # c.add_ref(
        #     hv_gen(c_inst=c_inst, volt=volt, dg_encx=dg_enc_cmp, dg_ency=dg_enc_poly)
        # )
        hv_gen(c, c_inst=c_inst, volt=volt, dg_encx=dg_enc_cmp, dg_ency=dg_enc_poly)

    elif bulk == "Bulk Tie":
        rect_bulk = c_inst.add_ref(
            gf.components.rectangle(
                size=(sd_l + con_sp, sd_diff.dysize), layer=layer["comp"]
            )
        )
        rect_bulk.dxmin = sd_diff.dxmax
        rect_bulk.dymin = sd_diff.dymin
        psdm = c_inst.add_ref(
            gf.components.rectangle(
                size=(
                    sd_diff.dxmax - sd_diff.dxmin + comp_pp_enc,
                    w_gate + 2 * gate_pp_enc,
                ),
                layer=layer["pplus"],
            )
        )
        psdm.dxmin = sd_diff.dxmin - comp_pp_enc
        psdm.dymin = sd_diff_intr.dymin - gate_pp_enc
        nsdm = c_inst.add_ref(
            gf.components.rectangle(
                size=(
                    rect_bulk.dxmax - rect_bulk.dxmin + comp_np_enc,
                    w_gate + 2 * comp_np_enc,
                ),
                layer=layer["nplus"],
            )
        )
        nsdm.connect("e1", psdm.ports["e3"])

        bulk_con = via_stack(
            x_range=(rect_bulk.dxmin + 0.1, rect_bulk.dxmax - 0.1),
            y_range=(rect_bulk.dymin, rect_bulk.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
        c_inst.add_ref(bulk_con)

        c.add_ref(c_inst)

        c.add_ref(
            labels_gen(
                label_str=sub_label,
                position=(
                    bulk_con.dxmin + bulk_con.dxsize / 2,
                    bulk_con.dymin + bulk_con.dysize / 2,
                ),
                layer=layer["metal1_label"],
                label=label,
                labels=[sub_label],
                label_valid_len=1,
            )
        )

        # deep nwell generation

        c.add_ref(
            pfet_deep_nwell(
                deepnwell=deepnwell,
                pcmpgr=pcmpgr,
                enc_size=(sd_diff.dxsize + rect_bulk.dxsize, sd_diff.dysize),
                enc_xmin=sd_diff.dxmin,
                enc_ymin=sd_diff.dymin,
                nw_enc_pcmp=nw_enc_pcmp,
                grw=grw,
            )
        )

        # dualgate generation
        # c.add_ref(
        #     hv_gen(c_inst=c_inst, volt=volt, dg_encx=dg_enc_cmp, dg_ency=dg_enc_poly)
        # )
        hv_gen(c, c_inst=c_inst, volt=volt, dg_encx=dg_enc_cmp, dg_ency=dg_enc_poly)

    elif bulk == "Guard Ring":
        psdm = c_inst.add_ref(
            gf.components.rectangle(
                size=(sd_diff.dxsize + 2 * comp_np_enc, w_gate + 2 * gate_pp_enc),
                layer=layer["pplus"],
            )
        )
        psdm.dxmin = sd_diff.dxmin - comp_pp_enc
        psdm.dymin = sd_diff_intr.dymin - gate_pp_enc
        c.add_ref(c_inst)

        bulk_gr_gen(
            c,
            c_inst=c_inst,
            comp_spacing=comp_spacing,
            poly2_comp_spacing=comp_spacing,
            volt=volt,
            grw=grw,
            l_d=l_d,
            implant_layer=layer["nplus"],
            label=label,
            sub_label=sub_label,
            deepnwell=deepnwell,
            pcmpgr=pcmpgr,
            nw_enc_pcmp=nw_enc_pcmp,
        )
        # bulk guardring

    return c


@gf.cell
def nfet_06v0_nvt(
    l_gate: float = 1.8,
    w_gate: float = 0.8,
    sd_con_col: int = 1,
    inter_sd_l: float = 0.24,
    nf: int = 1,
    grw: float = 0.22,
    bulk="None",
    con_bet_fin: int = 1,
    gate_con_pos="alternating",
    interdig: int = 0,
    patt="",
    label: bool = False,
    sd_label: Strs | None = [],
    g_label: str = [],
    sub_label: str = "",
    patt_label: bool = False,
) -> gf.Component:
    """Draw Native NFET 6V transistor by specifying parameters.

    Arg:
        l      : Float of gate length
        w      : Float of gate width
        ld     : Float of diffusion length
        nf     : Integer of number of fingers
        grw    : Float of guard ring width [If enabled]
        bulk   : String of bulk connection type [None, Bulk Tie, Guard Ring]
    """
    # used layers and dimensions
    end_cap: float = 0.22

    comp_spacing: float = 0.36
    poly2_comp_spacing: float = 0.3

    gate_np_enc: float = 0.23
    comp_np_enc: float = 0.16
    comp_pp_enc: float = 0.16
    poly2_spacing: float = 0.24
    pc_ext: float = 0.04

    con_size = 0.22
    con_sp = 0.28
    con_comp_enc = 0.07
    con_pl_enc = 0.07
    dg_enc_cmp = 0.24
    dg_enc_poly = 0.4

    sd_l_con = (
        ((sd_con_col) * con_size) + ((sd_con_col - 1) * con_sp) + 2 * con_comp_enc
    )
    sd_l = sd_l_con

    # gds components to store a single instance and the generated device
    c = gf.Component("sky_nfet_nvt_dev")

    c_inst = gf.Component("dev_temp")

    # generatingg sd diffusion

    if interdig == 1 and nf > 1 and nf != len(patt) and patt != "":
        nf = len(patt)

    l_d = (
        nf * l_gate + (nf - 1) * inter_sd_l + 2 * (con_comp_enc)
    )  # diffution total length
    rect_d_intr = gf.components.rectangle(size=(l_d, w_gate), layer=layer["comp"])
    sd_diff_intr = c_inst.add_ref(rect_d_intr)

    # generating sd contacts

    if w_gate <= con_size + 2 * con_comp_enc:
        cmpc_y = con_comp_enc + con_size + con_comp_enc

    else:
        cmpc_y = w_gate

    cmpc_size = (sd_l_con, cmpc_y)

    sd_diff = c_inst.add_ref(
        component=gf.components.rectangle(size=cmpc_size, layer=layer["comp"]),
        rows=1,
        columns=2,
        column_pitch=cmpc_size[0] + sd_diff_intr.dxsize,
    )

    sd_diff.dxmin = sd_diff_intr.dxmin - cmpc_size[0]
    sd_diff.dymin = sd_diff_intr.dymin - (sd_diff.dysize - sd_diff_intr.dysize) / 2

    sd_con = via_stack(
        x_range=(sd_diff.dxmin, sd_diff_intr.dxmin),
        y_range=(sd_diff.dymin, sd_diff.dymax),
        base_layer=layer["comp"],
        metal_level=1,
    )
    c_inst.add_ref(
        component=sd_con,
        columns=2,
        rows=1,
        column_pitch=sd_l + nf * l_gate + (nf - 1) * inter_sd_l + 2 * (con_comp_enc),
    )

    if con_bet_fin == 1 and nf > 1:
        inter_sd_con = via_stack(
            x_range=(
                sd_diff_intr.dxmin + con_comp_enc + l_gate,
                sd_diff_intr.dxmin + con_comp_enc + l_gate + inter_sd_l,
            ),
            y_range=(0, w_gate),
            base_layer=layer["comp"],
            metal_level=1,
        )
        c_inst.add_ref(
            component=inter_sd_con,
            columns=nf - 1,
            rows=1,
            column_pitch=l_gate + inter_sd_l,
        )

    ### adding source/drain labels
    c.add_ref(
        labels_gen(
            label_str="None",
            position=(
                sd_diff.dxmin + (sd_l / 2),
                sd_diff.dymin + (sd_diff.dysize / 2),
            ),
            layer=layer["metal1_label"],
            label=label,
            labels=sd_label,
            label_valid_len=nf + 1,
            index=0,
        )
    )

    c.add_ref(
        labels_gen(
            label_str="None",
            position=(
                sd_diff.dxmax - (sd_l / 2),
                sd_diff.dymin + (sd_diff.dysize / 2),
            ),
            layer=layer["metal1_label"],
            label=label,
            labels=sd_label,
            label_valid_len=nf + 1,
            index=nf,
        )
    )

    # generatingg poly

    if l_gate <= con_size + 2 * con_pl_enc:
        pc_x = con_pl_enc + con_size + con_pl_enc

    else:
        pc_x = l_gate

    pc_size = (pc_x, con_pl_enc + con_size + con_pl_enc)

    c_pc = gf.Component("poly con")

    rect_pc = c_pc.add_ref(gf.components.rectangle(size=pc_size, layer=layer["poly2"]))

    poly_con = via_stack(
        x_range=(rect_pc.dxmin, rect_pc.dxmax),
        y_range=(rect_pc.dymin, rect_pc.dymax),
        base_layer=layer["poly2"],
        metal_level=1,
        li_enc_dir="H",
    )
    c_pc.add_ref(poly_con)

    if nf == 1:
        poly = c_inst.add_ref(
            gf.components.rectangle(
                size=(l_gate, w_gate + 2 * end_cap), layer=layer["poly2"]
            )
        )
        poly.dxmin = sd_diff_intr.dxmin + con_comp_enc
        poly.dymin = sd_diff_intr.dymin - end_cap

        if gate_con_pos == "bottom":
            mv = 0
            nr = 1
        elif gate_con_pos == "top":
            mv = pc_size[1] + w_gate + 2 * end_cap
            nr = 1
        else:
            mv = 0
            nr = 2

        pc = c_inst.add_ref(
            component=c_pc,
            rows=nr,
            columns=1,
            row_pitch=pc_size[1] + w_gate + 2 * end_cap,
        )
        pc.dmove((poly.dxmin - ((pc_x - l_gate) / 2), -pc_size[1] - end_cap + mv))

        # gate_lablel
        c.add_ref(
            labels_gen(
                label_str="None",
                position=(pc.dxmin + c_pc.dxsize / 2, pc.dymin + c_pc.dysize / 2),
                layer=layer["metal1_label"],
                label=label,
                labels=g_label,
                label_valid_len=nf,
                index=0,
            )
        )

    else:
        w_p1 = end_cap + w_gate + end_cap  # poly total width

        if inter_sd_l < (poly2_spacing + 2 * pc_ext):
            if gate_con_pos == "alternating":
                w_p1 += 0.2
                w_p2 = w_p1
                e_c = 0.2
            else:
                w_p2 = w_p1 + con_pl_enc + con_size + con_pl_enc + poly2_spacing + 0.1
                e_c = 0

            if gate_con_pos == "bottom":
                p_mv = -end_cap - (w_p2 - w_p1)
            else:
                p_mv = -end_cap

        else:
            w_p2 = w_p1
            p_mv = -end_cap
            e_c = 0

        rect_p1 = gf.components.rectangle(size=(l_gate, w_p1), layer=layer["poly2"])
        rect_p2 = gf.components.rectangle(size=(l_gate, w_p2), layer=layer["poly2"])
        poly1 = c_inst.add_ref(
            rect_p1,
            rows=1,
            columns=ceil(nf / 2),
            column_pitch=2 * (inter_sd_l + l_gate),
        )
        poly1.dxmin = sd_diff_intr.dxmin + con_comp_enc
        poly1.dymin = sd_diff_intr.dymin - end_cap - e_c

        poly2 = c_inst.add_ref(
            rect_p2,
            rows=1,
            columns=floor(nf / 2),
            column_pitch=2 * (inter_sd_l + l_gate),
        )
        poly2.dxmin = poly1.dxmin + l_gate + inter_sd_l
        poly2.dymin = p_mv

        # generatingg poly contacts setups

        if gate_con_pos == "bottom":
            mv_1 = 0
            mv_2 = -(w_p2 - w_p1)
        elif gate_con_pos == "top":
            mv_1 = pc_size[1] + w_p1
            mv_2 = pc_size[1] + w_p2
        else:
            mv_1 = -e_c
            mv_2 = pc_size[1] + w_p2

        nc1 = ceil(nf / 2)
        nc2 = floor(nf / 2)

        pc_spacing = 2 * (inter_sd_l + l_gate)

        # generatingg poly contacts

        pc1 = c_inst.add_ref(
            component=c_pc, rows=1, columns=nc1, column_pitch=pc_spacing
        )
        pc1.dmove((poly1.dxmin - ((pc_x - l_gate) / 2), -pc_size[1] - end_cap + mv_1))

        pc2 = c_inst.add_ref(
            component=c_pc, rows=1, columns=nc2, column_pitch=pc_spacing
        )
        pc2.dmove(
            (
                poly1.dxmin - ((pc_x - l_gate) / 2) + (inter_sd_l + l_gate),
                -pc_size[1] - end_cap + mv_2,
            )
        )

        add_inter_sd_labels(
            c,
            nf,
            sd_label,
            poly1,
            l_gate,
            inter_sd_l,
            sd_diff_intr,
            label,
            layer,
            con_bet_fin,
        )

        add_gate_labels(
            c, g_label, pc1, c_pc, pc_spacing, nc1, nc2, pc2, label, layer, nf
        )

        if interdig == 1:
            c_inst.add_ref(
                interdigit(
                    sd_diff=sd_diff,
                    pc1=pc1,
                    pc2=pc2,
                    poly_con=poly_con,
                    sd_l=sd_l,
                    nf=nf,
                    patt=patt,
                    gate_con_pos=gate_con_pos,
                    pc_x=pc_x,
                    pc_spacing=pc_spacing,
                    label=label,
                    g_label=g_label,
                    patt_label=patt_label,
                )
            )

    # generatingg bulk
    if bulk == "None":
        nplus = c_inst.add_ref(
            gf.components.rectangle(
                size=(sd_diff.dxsize + 2 * comp_np_enc, w_gate + 2 * gate_np_enc),
                layer=layer["nplus"],
            )
        )
        nplus.dxmin = sd_diff.dxmin - comp_np_enc
        nplus.dymin = sd_diff_intr.dymin - gate_np_enc

    elif bulk == "Bulk Tie":
        rect_bulk = c_inst.add_ref(
            gf.components.rectangle(
                size=(sd_l + con_sp, sd_diff.dysize), layer=layer["comp"]
            )
        )
        rect_bulk.dxmin = sd_diff.dxmax
        rect_bulk.dymin = sd_diff.dymin
        nsdm = c_inst.add_ref(
            gf.components.rectangle(
                size=(
                    sd_diff.dxmax - sd_diff.dxmin + comp_np_enc,
                    w_gate + 2 * gate_np_enc,
                ),
                layer=layer["nplus"],
            )
        )
        nsdm.dxmin = sd_diff.dxmin - comp_np_enc
        nsdm.dymin = sd_diff_intr.dymin - gate_np_enc
        psdm = c_inst.add_ref(
            gf.components.rectangle(
                size=(
                    rect_bulk.dxmax - rect_bulk.dxmin + comp_pp_enc,
                    w_gate + 2 * comp_pp_enc,
                ),
                layer=layer["pplus"],
            )
        )
        psdm.connect("e1", nsdm.ports["e3"])

        bulk_con = via_stack(
            x_range=(rect_bulk.dxmin + 0.1, rect_bulk.dxmax - 0.1),
            y_range=(rect_bulk.dymin, rect_bulk.dymax),
            base_layer=layer["comp"],
            metal_level=1,
        )
        c_inst.add_ref(bulk_con)

        c.add_ref(
            labels_gen(
                label_str=sub_label,
                position=(
                    bulk_con.dxmin + bulk_con.dxsize / 2,
                    bulk_con.dymin + bulk_con.dysize / 2,
                ),
                layer=layer["metal1_label"],
                label=label,
                labels=[sub_label],
                label_valid_len=1,
            )
        )

    elif bulk == "Guard Ring":
        nsdm = c_inst.add_ref(
            gf.components.rectangle(
                size=(sd_diff.dxsize + 2 * comp_np_enc, w_gate + 2 * gate_np_enc),
                layer=layer["nplus"],
            )
        )
        nsdm.dxmin = sd_diff.dxmin - comp_np_enc
        nsdm.dymin = sd_diff_intr.dymin - gate_np_enc
        c.add_ref(c_inst)

        c_temp = gf.Component("temp_store")
        rect_bulk_in = c_temp.add_ref(
            gf.components.rectangle(
                size=(
                    (c_inst.dxmax - c_inst.dxmin) + 2 * comp_spacing,
                    (c_inst.dymax - c_inst.dymin) + 2 * poly2_comp_spacing,
                ),
                layer=layer["comp"],
            )
        )
        rect_bulk_in.dmove(
            (c_inst.dxmin - comp_spacing, c_inst.dymin - poly2_comp_spacing)
        )
        rect_bulk_out = c_temp.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_bulk_in.dxmax - rect_bulk_in.dxmin) + 2 * grw,
                    (rect_bulk_in.dymax - rect_bulk_in.dymin) + 2 * grw,
                ),
                layer=layer["comp"],
            )
        )
        rect_bulk_out.dmove((rect_bulk_in.dxmin - grw, rect_bulk_in.dymin - grw))
        B = c.add_ref(
            gf.boolean(
                A=rect_bulk_out,
                B=rect_bulk_in,
                operation="A-B",
                layer=layer["comp"],
            )
        )

        psdm_in = c_temp.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_bulk_in.dxmax - rect_bulk_in.dxmin) - 2 * comp_pp_enc,
                    (rect_bulk_in.dymax - rect_bulk_in.dymin) - 2 * comp_pp_enc,
                ),
                layer=layer["pplus"],
            )
        )
        psdm_in.dmove(
            (rect_bulk_in.dxmin + comp_pp_enc, rect_bulk_in.dymin + comp_pp_enc)
        )
        psdm_out = c_temp.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_bulk_out.dxmax - rect_bulk_out.dxmin) + 2 * comp_pp_enc,
                    (rect_bulk_out.dymax - rect_bulk_out.dymin) + 2 * comp_pp_enc,
                ),
                layer=layer["pplus"],
            )
        )
        psdm_out.dmove(
            (
                rect_bulk_out.dxmin - comp_pp_enc,
                rect_bulk_out.dymin - comp_pp_enc,
            )
        )
        psdm = c.add_ref(
            gf.boolean(A=psdm_out, B=psdm_in, operation="A-B", layer=layer["pplus"])
        )

        # generatingg contacts

        c.add_ref(
            via_generator(
                x_range=(
                    rect_bulk_in.dxmin + con_size,
                    rect_bulk_in.dxmax - con_size,
                ),
                y_range=(rect_bulk_out.dymin, rect_bulk_in.dymin),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # bottom contact

        c.add_ref(
            via_generator(
                x_range=(
                    rect_bulk_in.dxmin + con_size,
                    rect_bulk_in.dxmax - con_size,
                ),
                y_range=(rect_bulk_in.dymax, rect_bulk_out.dymax),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # upper contact

        c.add_ref(
            via_generator(
                x_range=(rect_bulk_out.dxmin, rect_bulk_in.dxmin),
                y_range=(
                    rect_bulk_in.dymin + con_size,
                    rect_bulk_in.dymax - con_size,
                ),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # right contact

        c.add_ref(
            via_generator(
                x_range=(rect_bulk_in.dxmax, rect_bulk_out.dxmax),
                y_range=(
                    rect_bulk_in.dymin + con_size,
                    rect_bulk_in.dymax - con_size,
                ),
                via_enclosure=(con_comp_enc, con_comp_enc),
                via_layer=layer["contact"],
                via_size=(con_size, con_size),
                via_spacing=(con_sp, con_sp),
            )
        )  # left contact

        comp_m1_in = c_temp.add_ref(
            gf.components.rectangle(
                size=(
                    (l_d) + 2 * comp_spacing,
                    (c_inst.dymax - c_inst.dymin) + 2 * poly2_comp_spacing,
                ),
                layer=layer["metal1"],
            )
        )
        comp_m1_in.dmove((-comp_spacing, c_inst.dymin - poly2_comp_spacing))
        comp_m1_out = c_temp.add_ref(
            gf.components.rectangle(
                size=(
                    (rect_bulk_in.dxmax - rect_bulk_in.dxmin) + 2 * grw,
                    (rect_bulk_in.dymax - rect_bulk_in.dymin) + 2 * grw,
                ),
                layer=layer["metal1"],
            )
        )
        comp_m1_out.dmove((rect_bulk_in.dxmin - grw, rect_bulk_in.dymin - grw))
        b_gr = c.add_ref(
            gf.boolean(
                A=rect_bulk_out,
                B=rect_bulk_in,
                operation="A-B",
                layer=layer["metal1"],
            )
        )  # guardring metal1

        c.add_ref(
            labels_gen(
                label_str=sub_label,
                position=(
                    b_gr.dxmin + (grw + 2 * (comp_pp_enc)) / 2,
                    b_gr.dymin + (b_gr.dysize / 2),
                ),
                layer=layer["metal1_label"],
                label=label,
                labels=[sub_label],
                label_valid_len=1,
            )
        )

        dg = c.add_ref(
            gf.components.rectangle(
                size=(
                    B.dxsize + (2 * dg_enc_cmp),
                    B.dysize + (2 * dg_enc_cmp),
                ),
                layer=layer["dualgate"],
            )
        )
        dg.dxmin = B.dxmin - dg_enc_cmp
        dg.dymin = B.dymin - dg_enc_cmp

    if bulk != "Guard Ring":
        c.add_ref(c_inst)

        dg = c.add_ref(
            gf.components.rectangle(
                size=(
                    c_inst.dxsize + (2 * dg_enc_cmp),
                    c_inst.dysize + (2 * dg_enc_poly),
                ),
                layer=layer["dualgate"],
            )
        )
        dg.dxmin = c_inst.dxmin - dg_enc_cmp
        dg.dymin = c_inst.dymin - dg_enc_poly

    # generatingg native layer
    nat = c.add_ref(
        gf.components.rectangle(size=(dg.dxsize, dg.dysize), layer=layer["nat"])
    )

    nat.dxmin = dg.dxmin
    nat.dymin = dg.dymin

    return c


if __name__ == "__main__":
    c = pfet()
    # c = nfet()
    # c = nfet_06v0_nvt()
    # c = interdigit()
    # c = alter_interdig()
    c.show()



import gdsfactory as gf
from gdsfactory.technology import LayerLevel, LayerMap, LayerStack
from gdsfactory.typings import Layer

from gf180mcu.config import PATH


class LAYER(LayerMap):
    comp: Layer = (22, 0)
    dnwell: Layer = (12, 0)
    nwell: Layer = (21, 0)
    lvpwell: Layer = (204, 0)
    dualgate: Layer = (55, 0)
    poly2: Layer = (30, 0)
    nplus: Layer = (32, 0)
    pplus: Layer = (31, 0)
    sab: Layer = (49, 0)
    esd: Layer = (24, 0)
    contact: Layer = (33, 0)
    metal1: Layer = (34, 0)
    via1: Layer = (35, 0)
    metal2: Layer = (36, 0)
    via2: Layer = (38, 0)
    metal3: Layer = (42, 0)
    via3: Layer = (40, 0)
    metal4: Layer = (46, 0)
    via4: Layer = (41, 0)
    metal5: Layer = (81, 0)
    via5: Layer = (82, 0)
    metaltop: Layer = (53, 0)
    pad: Layer = (37, 0)
    resistor: Layer = (62, 0)
    fhres: Layer = (227, 0)
    fusetop: Layer = (75, 0)
    fusewindow_d: Layer = (96, 1)
    polyfuse: Layer = (220, 0)
    mvsd: Layer = (210, 0)
    mvpsd: Layer = (11, 39)
    nat: Layer = (5, 0)
    comp_dummy: Layer = (22, 4)
    poly2_dummy: Layer = (30, 4)
    metal1_dummy: Layer = (34, 4)
    metal2_dummy: Layer = (36, 4)
    metal3_dummy: Layer = (42, 4)
    metal4_dummy: Layer = (46, 4)
    metal5_dummy: Layer = (81, 4)
    metaltop_dummy: Layer = (53, 4)
    comp_label: Layer = (22, 10)
    poly2_label: Layer = (30, 10)
    metal1_label: Layer = (34, 10)
    metal2_label: Layer = (36, 10)
    metal3_label: Layer = (42, 10)
    metal4_label: Layer = (46, 10)
    metal5_label: Layer = (81, 10)
    metaltop_label: Layer = (53, 10)
    metal1_slot: Layer = (34, 3)
    metal2_slot: Layer = (36, 3)
    metal3_slot: Layer = (42, 3)
    metal4_slot: Layer = (46, 3)
    metal5_slot: Layer = (81, 3)
    metaltop_slot: Layer = (53, 3)
    ubmpperi: Layer = (183, 0)
    ubmparray: Layer = (184, 0)
    ubmeplate: Layer = (185, 0)
    schottky_diode: Layer = (241, 0)
    zener: Layer = (178, 0)
    res_mk: Layer = (110, 5)
    opc_drc: Layer = (124, 5)
    ndmy: Layer = (111, 5)
    pmndmy: Layer = (152, 5)
    v5_xtor: Layer = (112, 1)
    cap_mk: Layer = (117, 5)
    mos_cap_mk: Layer = (166, 5)
    ind_mk: Layer = (151, 5)
    diode_mk: Layer = (115, 5)
    drc_bjt: Layer = (127, 5)
    lvs_bjt: Layer = (118, 5)
    mim_l_mk: Layer = (117, 10)
    latchup_mk: Layer = (137, 5)
    guard_ring_mk: Layer = (167, 5)
    otp_mk: Layer = (173, 5)
    mtpmark: Layer = (122, 5)
    neo_ee_mk: Layer = (88, 17)
    sramcore: Layer = (108, 5)
    lvs_rf: Layer = (100, 5)
    lvs_drain: Layer = (100, 7)
    ind_mk1: Layer = (151, 5)
    hvpolyrs: Layer = (123, 5)
    lvs_io: Layer = (119, 5)
    probe_mk: Layer = (13, 17)
    esd_mk: Layer = (24, 5)
    lvs_source: Layer = (100, 8)
    well_diode_mk: Layer = (153, 51)
    ldmos_xtor: Layer = (226, 0)
    plfuse: Layer = (125, 5)
    efuse_mk: Layer = (80, 5)
    mcell_feol_mk: Layer = (11, 17)
    ymtp_mk: Layer = (86, 17)
    dev_wf_mk: Layer = (128, 17)
    metal1_blk: Layer = (34, 5)
    metal2_blk: Layer = (36, 5)
    metal3_blk: Layer = (42, 5)
    metal4_blk: Layer = (46, 5)
    metal5_blk: Layer = (81, 5)
    metalt_blk: Layer = (53, 5)
    pr_bndry: Layer = (0, 0)
    mdiode: Layer = (116, 5)
    metal1_res: Layer = (110, 11)
    metal2_res: Layer = (110, 12)
    metal3_res: Layer = (110, 13)
    metal4_res: Layer = (110, 14)
    metal5_res: Layer = (110, 15)
    metal6_res: Layer = (110, 16)
    border: Layer = (63, 0)


layer = LAYER
LAYER_VIEWS = gf.technology.LayerViews(PATH.lyp_yaml)


# Define layer stack for GF180MCU 180nm process
# Based on typical 180nm CMOS process parameters
def get_layer_stack(
    thickness_substrate: float = 200.0,  # substrate thickness in um
    thickness_dnwell: float = 2.5,  # deep n-well thickness in um
    thickness_nwell: float = 1.8,  # n-well thickness in um
    thickness_comp: float = 0.18,  # active area thickness in um
    thickness_poly2: float = 0.18,  # poly2 thickness in um
    thickness_contact: float = 0.4,  # contact thickness in um
    thickness_metal1: float = 0.35,  # metal1 thickness in um
    thickness_via1: float = 0.4,  # via1 thickness in um
    thickness_metal2: float = 0.35,  # metal2 thickness in um
    thickness_via2: float = 0.4,  # via2 thickness in um
    thickness_metal3: float = 0.7,  # metal3 thickness in um
    thickness_via3: float = 0.4,  # via3 thickness in um
    thickness_metal4: float = 0.7,  # metal4 thickness in um
    thickness_via4: float = 0.4,  # via4 thickness in um
    thickness_metal5: float = 0.7,  # metal5 thickness in um
    thickness_via5: float = 0.4,  # via5 thickness in um
    thickness_metaltop: float = 2.0,  # top metal thickness in um
    thickness_pad: float = 10.0,  # pad thickness in um
) -> LayerStack:
    """Return GF180MCU LayerStack.

    Based on GlobalFoundries 180nm MCU process technology.
    Typical layer thicknesses for 180nm CMOS process.

    The LayerStack defines the physical properties of each layer including:
    - Layer thickness
    - Z-position in the stack
    - Material properties
    - Mesh order for simulations

    Usage:
        ```python
        import gf180mcu

        # Use default layer stack
        layer_stack = gf180mcu.LAYER_STACK

        # Create custom layer stack with different parameters
        custom_stack = gf180mcu.get_layer_stack(
            thickness_metal1=0.5,  # thicker metal1
            thickness_metaltop=3.0  # thicker top metal
        )

        # Use for 3D visualization
        import gdsfactory as gf
        c = gf.components.rectangle(layer=gf180mcu.LAYER.metal1)
        scene = c.to_3d(layer_stack=layer_stack)
        ```

    Args:
        thickness_substrate: substrate thickness
        thickness_dnwell: deep n-well thickness
        thickness_nwell: n-well thickness
        thickness_comp: active area thickness
        thickness_poly2: poly2 thickness
        thickness_contact: contact thickness
        thickness_metal1: metal1 thickness
        thickness_via1: via1 thickness
        thickness_metal2: metal2 thickness
        thickness_via2: via2 thickness
        thickness_metal3: metal3 thickness
        thickness_via3: via3 thickness
        thickness_metal4: metal4 thickness
        thickness_via4: via4 thickness
        thickness_metal5: metal5 thickness
        thickness_via5: via5 thickness
        thickness_metaltop: top metal thickness
        thickness_pad: pad thickness

    Returns:
        LayerStack: Complete layer stack for GF180MCU process
    """

    # Z-positions (cumulative heights)
    z_substrate = 0.0
    z_dnwell = z_substrate
    z_nwell = z_substrate
    z_comp = z_substrate
    z_poly2 = z_substrate
    z_contact = z_substrate + thickness_comp + thickness_poly2
    z_metal1 = z_contact + thickness_contact
    z_via1 = z_metal1 + thickness_metal1
    z_metal2 = z_via1 + thickness_via1
    z_via2 = z_metal2 + thickness_metal2
    z_metal3 = z_via2 + thickness_via2
    z_via3 = z_metal3 + thickness_metal3
    z_metal4 = z_via3 + thickness_via3
    z_via4 = z_metal4 + thickness_metal4
    z_metal5 = z_via4 + thickness_via4
    z_via5 = z_metal5 + thickness_metal5
    z_metaltop = z_via5 + thickness_via5
    z_pad = z_metaltop

    layers = {
        # Substrate and wells
        "substrate": LayerLevel(
            layer=LAYER.pr_bndry,
            thickness=thickness_substrate,
            zmin=z_substrate - thickness_substrate,
            material="si",
            mesh_order=100,
        ),
        "dnwell": LayerLevel(
            layer=LAYER.dnwell,
            thickness=thickness_dnwell,
            zmin=z_dnwell,
            material="si",
            mesh_order=10,
        ),
        "nwell": LayerLevel(
            layer=LAYER.nwell,
            thickness=thickness_nwell,
            zmin=z_nwell,
            material="si",
            mesh_order=9,
        ),
        "lvpwell": LayerLevel(
            layer=LAYER.lvpwell,
            thickness=thickness_nwell,  # similar to nwell
            zmin=z_nwell,
            material="si",
            mesh_order=9,
        ),
        # Active areas and implants
        "comp": LayerLevel(
            layer=LAYER.comp,
            thickness=thickness_comp,
            zmin=z_comp,
            material="si",
            mesh_order=8,
        ),
        "nplus": LayerLevel(
            layer=LAYER.nplus,
            thickness=thickness_comp,
            zmin=z_comp,
            material="si",
            mesh_order=7,
        ),
        "pplus": LayerLevel(
            layer=LAYER.pplus,
            thickness=thickness_comp,
            zmin=z_comp,
            material="si",
            mesh_order=7,
        ),
        "nat": LayerLevel(
            layer=LAYER.nat,
            thickness=thickness_comp,
            zmin=z_comp,
            material="si",
            mesh_order=7,
        ),
        # Polysilicon
        "poly2": LayerLevel(
            layer=LAYER.poly2,
            thickness=thickness_poly2,
            zmin=z_poly2,
            material="poly",
            mesh_order=6,
        ),
        # Contacts and vias
        "contact": LayerLevel(
            layer=LAYER.contact,
            thickness=thickness_contact,
            zmin=z_contact,
            material="tungsten",
            mesh_order=5,
        ),
        "via1": LayerLevel(
            layer=LAYER.via1,
            thickness=thickness_via1,
            zmin=z_via1,
            material="tungsten",
            mesh_order=5,
        ),
        "via2": LayerLevel(
            layer=LAYER.via2,
            thickness=thickness_via2,
            zmin=z_via2,
            material="tungsten",
            mesh_order=5,
        ),
        "via3": LayerLevel(
            layer=LAYER.via3,
            thickness=thickness_via3,
            zmin=z_via3,
            material="tungsten",
            mesh_order=5,
        ),
        "via4": LayerLevel(
            layer=LAYER.via4,
            thickness=thickness_via4,
            zmin=z_via4,
            material="tungsten",
            mesh_order=5,
        ),
        "via5": LayerLevel(
            layer=LAYER.via5,
            thickness=thickness_via5,
            zmin=z_via5,
            material="tungsten",
            mesh_order=5,
        ),
        # Metal layers
        "metal1": LayerLevel(
            layer=LAYER.metal1,
            thickness=thickness_metal1,
            zmin=z_metal1,
            material="aluminum",
            mesh_order=4,
        ),
        "metal2": LayerLevel(
            layer=LAYER.metal2,
            thickness=thickness_metal2,
            zmin=z_metal2,
            material="aluminum",
            mesh_order=4,
        ),
        "metal3": LayerLevel(
            layer=LAYER.metal3,
            thickness=thickness_metal3,
            zmin=z_metal3,
            material="aluminum",
            mesh_order=4,
        ),
        "metal4": LayerLevel(
            layer=LAYER.metal4,
            thickness=thickness_metal4,
            zmin=z_metal4,
            material="aluminum",
            mesh_order=4,
        ),
        "metal5": LayerLevel(
            layer=LAYER.metal5,
            thickness=thickness_metal5,
            zmin=z_metal5,
            material="aluminum",
            mesh_order=4,
        ),
        "metaltop": LayerLevel(
            layer=LAYER.metaltop,
            thickness=thickness_metaltop,
            zmin=z_metaltop,
            material="aluminum",
            mesh_order=4,
        ),
        # Pads and special layers
        "pad": LayerLevel(
            layer=LAYER.pad,
            thickness=thickness_pad,
            zmin=z_pad,
            material="aluminum",
            mesh_order=3,
        ),
        # Resistors
        "resistor": LayerLevel(
            layer=LAYER.resistor,
            thickness=thickness_poly2,  # typically same as poly
            zmin=z_poly2,
            material="poly",
            mesh_order=6,
        ),
        "fhres": LayerLevel(
            layer=LAYER.fhres,
            thickness=thickness_poly2,
            zmin=z_poly2,
            material="poly",
            mesh_order=6,
        ),
        # Special device layers (markers, typically no physical thickness)
        "sab": LayerLevel(
            layer=LAYER.sab,
            thickness=0.001,  # minimal thickness for visualization
            zmin=z_comp,
            material="si",
            mesh_order=1,
        ),
        "esd": LayerLevel(
            layer=LAYER.esd,
            thickness=0.001,
            zmin=z_comp,
            material="si",
            mesh_order=1,
        ),
    }

    return LayerStack(layers=layers)


# Create default layer stack instance
LAYER_STACK = get_layer_stack()

if __name__ == "__main__":
    # LAYER_VIEWS.to_lyp(PATH.lyp)

    # Demo: LayerStack usage
    print("GF180MCU LayerStack Demo")
    print("=" * 40)
    print(f"Number of layers: {len(LAYER_STACK.layers)}")
    print(
        f"Total stack height: {max(layer.zmin + layer.thickness for layer in LAYER_STACK.layers.values()):.2f} um"
    )
    print("\nMetal layers:")
    for name, layer in LAYER_STACK.layers.items():
        if "metal" in name:
            print(
                f"  {name}: {layer.thickness}um thick at z={layer.zmin:.2f}um ({layer.material})"
            )

    print("\nCustom layer stack example:")
    custom = get_layer_stack(thickness_metal1=0.5, thickness_metaltop=3.0)
    print(f"  Custom metal1: {custom.layers['metal1'].thickness}um")
    print(f"  Custom metaltop: {custom.layers['metaltop'].thickness}um")



import sys

from gdsfactory.get_factories import get_cells
from gdsfactory.pdk import Pdk

from gf180mcu import config, diode, fet, layers
from gf180mcu.cap_mim import (
    cap_mim,
)
from gf180mcu.cap_mos import (
    cap_mos,
    cap_mos_inst,
)
from gf180mcu.config import PATH
from gf180mcu.diode import (
    diode_dw2ps,
    diode_nd2ps,
    diode_nw2ps,
    diode_pd2nw,
    diode_pw2dw,
    sc_diode,
)
from gf180mcu.fet import (
    add_gate_labels,
    add_inter_sd_labels,
    alter_interdig,
    bulk_gr_gen,
    get_patt_label,
    hv_gen,
    interdigit,
    labels_gen,
    nfet,
    nfet_06v0_nvt,
    nfet_deep_nwell,
    pfet,
    pfet_deep_nwell,
)
from gf180mcu.guardring import (
    pcmpgr_gen,
)
from gf180mcu.layers import (
    LAYER,
    LAYER_STACK,
    LAYER_VIEWS,
    LayerMap,
    get_layer_stack,
    layer,
)
from gf180mcu.res import (
    nplus_res,
    npolyf_res,
    plus_res_inst,
    polyf_res_inst,
    pplus_res,
    ppolyf_res,
    ppolyf_u_high_Rs_res,
    res,
    well_res,
)
from gf180mcu.via_generator import (
    via_generator,
    via_stack,
)

__all__ = [
    "LAYER",
    "LAYER_STACK",
    "LAYER_VIEWS",
    "PATH",
    "LayerMap",
    "add_gate_labels",
    "add_inter_sd_labels",
    "alter_interdig",
    "bulk_gr_gen",
    "cap_mim",
    "cap_mos",
    "cap_mos",
    "cap_mos_inst",
    "config",
    "diode",
    "diode_dw2ps",
    "diode_nd2ps",
    "diode_nw2ps",
    "diode_pd2nw",
    "diode_pw2dw",
    "fet",
    "get_layer_stack",
    "get_patt_label",
    "hv_gen",
    "interdigit",
    "labels_gen",
    "layer",
    "layers",
    "nfet",
    "nfet_06v0_nvt",
    "nfet_deep_nwell",
    "nplus_res",
    "npolyf_res",
    "pcmpgr_gen",
    "pfet",
    "pfet_deep_nwell",
    "plus_res_inst",
    "polyf_res_inst",
    "pplus_res",
    "ppolyf_res",
    "ppolyf_u_high_Rs_res",
    "res",
    "sc_diode",
    "via_generator",
    "via_stack",
    "well_res",
]
__version__ = "0.3.0"

cells = get_cells(sys.modules[__name__])
PDK = Pdk(
    name="gf180mcu",
    cells=cells,
    layers=LAYER,
    layer_views=LAYER_VIEWS,
    layer_stack=LAYER_STACK,
)
PDK.activate()
